import os
import json
import numpy as np
import pickle

from collections import OrderedDict
from datetime import datetime as dt

class _BasicTFKerasMixin(object):
  """
  Mixin for basic tf keras functionalities that are attached to `libraries.logger.Logger`.

  This mixin cannot be instantiated because it is built just to provide some additional
  functionalities for `libraries.logger.Logger`

  In this mixin we can use any attribute/method of the Logger.

  * Obs: This mixin uses also attributes/methods of `_JSONSerializationMixin`"
    - self.save_json

  * Obs: This mixin uses also attributes/methods of `_MatplotlibMixin`:
    - self.add_copyright_to_plot
    - self.output_pyplot_image

  * Obs: This mixin uses also attributess/mthods of `_PublicTFKerasMixin`:
    - self._check_tf_avail
  """

  def __init__(self):
    super(_BasicTFKerasMixin, self).__init__()

    try:
      from ratio1.logging.logger_mixins.pickle_serialization_mixin import _PickleSerializationMixin
    except ModuleNotFoundError:
      raise ModuleNotFoundError("Cannot use _BasicTFKerasMixin without having _PickleSerializationMixin")

    try:
      from .matplotlib_mixin import _MatplotlibMixin
    except ModuleNotFoundError:
      raise ModuleNotFoundError("Cannot use _BasicTFKerasMixin without having _MatplotlibMixin")

    # self.gpu_mem = None
    # self.TF = False
    # self.TF_VER = None
    # self.TF_KERAS_VER = None
    # self.KERAS = False
    # self.KERAS_VER = None
    # self.devices = {}
    self._model_save_history = []
    self._model_save_last = []
    return

  def initializers(self, seed=123):
    import tensorflow as tf
    self._check_tf_avail()

    dct = {
      'zeros': tf.keras.initializers.Zeros(),
      'glorot_uniform': tf.keras.initializers.glorot_uniform(seed=seed),
      'uniform': tf.keras.initializers.RandomUniform(seed=seed),
      'orthogonal': tf.keras.initializers.orthogonal(seed=seed)
    }
    return dct

  def plot_keras_model(self, model, label='', verbose=True):
    """
    plots a model or generates a model text summary file
    """
    from tensorflow.keras.utils import plot_model
    if label == '':
      label = model.name
    _fn = os.path.join(self.get_output_folder(), label)
    if verbose:
      self.P("Plotting model '{}'...".format(model.name))
    try:
      fn = _fn + '_graph.png'
      res = plot_model(
        model,
        to_file=fn,
        show_shapes=True,
        show_layer_names=True,
        )
      if verbose and res is not None:
        self.P("  Model plotted and saved to '{}'.".format(fn[-40:]))
      if res is None:
        self.P("ERROR: plot_model probably failed!")
    except:
      fn = _fn + '.txt'
      s_model = self.get_keras_model_summary(model)
      with open(fn, 'wt') as f:
        f.write(s_model)
      self.P("  Model plot failed. Saved model descr in '{}'".format(fn[-40:]))
      self.P("  Please use 'conda install -c anaconda pydot' in order to plot models")
    return

  @staticmethod
  def get_keras_model_desc(model):
    """
    gets keras model short description
    """
    short_name = ""
    nr_l = len(model.layers)
    for i in range(nr_l):
      layer = model.layers[i]
      s_layer = "{}".format(layer.name)
      c_layer = s_layer.upper()[0:4]
      if c_layer == "CONV":
        c_layer = "Conv{}".format(layer.filters)

      if c_layer == "DENS":
        c_layer = "DNS{}".format(layer.units)

      if c_layer == "DROP":
        c_layer = "DRP"

      c_layer += "+"

      short_name += c_layer
    short_name = short_name[:-1]
    return short_name

  def save_model_notes(self, l_notes, model_name, cfg, DEBUG=False):
    if DEBUG:
      self.P("  Saving {} notes for {}".format(len(l_notes), model_name))
    fn = os.path.join(self.get_models_folder(), model_name + '.txt')
    with open(fn, "w") as fp:
      fp.write("Model: {} [{}]\n\n".format(
        model_name, dt.now().strftime("%Y-%m-%d %H:%M:%S")))
      for _l in l_notes:
        fp.write("{}\n".format(_l))
    self._save_model_config(model_name, cfg)
    return

  def _save_model_config(self, label, cfg, DEBUG=False):
    if cfg is not None:
      if type(cfg) is bool:
        cfg = self.config_data
      elif type(cfg) is dict:
        pass
      cfg['INFERENCE'] = label
      if DEBUG:
        self.verbose_log("  Saving cfg [{}] to models...".format(label))
      file_path = os.path.join(self.get_models_folder(), label + '.json')
      with open(file_path, 'w') as fp:
        json.dump(cfg, fp, sort_keys=False, indent=4)
    return

  def delete_model(self, model_name):
    return self.remove_model(model_name)

  def remove_model(self, model_name):
    if model_name[-3:].lower() != '.h5':
      model_name += '.h5'
    return self._delete_h5_model(model_name)

  def _delete_h5_model(self, _fn):
    folder = self.get_models_folder()
    fn = os.path.join(folder, _fn)
    result = False
    if os.path.isfile(fn):
      if '.h5' in fn:
        try:
          os.remove(fn)
          self.P("  Model file '{}' deleted!".format(fn[-50:]))
          result = True
        except:
          pass
    else:
      result = True  # file does not exist anyway
    return result

  def _delete_model_file(self, _fn):
    folder = self.get_models_folder()
    fn = os.path.join(folder, _fn)
    result = False
    if os.path.isfile(fn):
      try:
        os.remove(fn)
        self.P("  File '{}' deleted!".format(fn[-50:]))
        result = True
      except:
        pass
    else:
      result = True  # file does not exist anyway
    return result

  def clean_all_h5(self):
    folder = self.get_models_folder()
    all_files = os.listdir(folder)
    errs = []
    self.P("Cleaning all .h5 files from {}".format(folder))

    for _fn in all_files:
      if not self._delete_h5_model(_fn):
        errs.append(_fn)

    for fn in errs:
      if not self._delete_h5_model(fn):
        self.P("  ERROR: file '{}' could not be deleted".format(fn[-50:]))

    return

  def save_keras_model_def_and_weights(self, model, label, subfolder_path=None):
    """
    saves the definition of the keras model, as well as all the weights
    """

    save_folder = self.get_models_folder()
    if subfolder_path is not None:
      save_folder = os.path.join(save_folder, subfolder_path.lstrip('/'))

    if not os.path.exists(save_folder):
      os.makedirs(save_folder)

    path_model_json = os.path.join(save_folder, label + '_model_def.json')
    path_model_weights = os.path.join(save_folder, label + '_weights.h5')

    model_json = json.loads(model.to_json())
    self.save_json(model_json, path_model_json)
    self.P("Saved model json   : '...{}'".format(path_model_json[-40:]))

    model.save_weights(filepath=path_model_weights)
    self.P("Saved model weights: '...{}'".format(path_model_weights[-40:]))

    return

  def save_keras_model_weights(self, filename, model, layers, ):
    """
    save the weights of 'layers' list in file
    """
    assert len(layers) > 0, 'Unknown list of selected layers'
    file_name = os.path.join(self.get_models_folder(), filename + ".pkl")
    self.P("Saving weights for {} layers in {}...".format(len(layers), file_name))
    w_dict = OrderedDict()
    for layer in layers:
      w_dict[layer] = model.get_layer(layer).get_weights()
    with open(file_name, 'wb') as f:
      pickle.dump(w_dict, f)
    self.P("Done saving weights [{}].".format(layers), show_time=True)
    return

  def reset_saved_models_history(self):
    done = False
    while not done:
      if len(self._model_save_history) > 0:
        self._delete_model_history()
      else:
        done = True
    self._model_save_history = []
    self._model_save_last = []
    return

  def _delete_model_history(self):
    deleted_files = []
    for fn in self._model_save_history:
      if os.path.isfile(fn):
        try:
          os.remove(fn)
          deleted_files.append(fn)
        except:
          self.P("    Error deleting {}".format(fn))
      else:
        deleted_files.append(fn)
    self._model_save_history = [x for x in self._model_save_history if x not in deleted_files]
    return

  def save_keras_model(self, model, label, use_prefix=False,
                       use_single_prefix=False,
                       cfg=None,
                       include_optimizer=True, DEBUG=True,
                       only_weights=False,
                       delete_previous=False, record_previous=False,
                       save_model_and_weights=False,
                       to_models=True,
                       subfolder_path=None):
    """
    saves keras model to a file

    Parameters:
    -----------
    model : keras model
    label : name of the model
    use_prefix : will add date-time prefix if True (default False)
    use_single_prefix :  will use only one timestamp prefix for all saved models
    save_model_and_weights :  True if you want to save them both!

    subfolder_path : str, optional
      A path relative to '_models' or '_output' (_output if `to_models=False`) where the model is saved
      Default is None.
    """
    save_model = (not only_weights) or save_model_and_weights
    save_weights = only_weights or save_model_and_weights

    MDL_EXT = '.h5'
    WGH_EXT = '.h5'
    file_prefix = ""
    if label == "":
      label = self.get_keras_model_desc(model)
    label = label.replace(">", "_")  # [:30]

    if use_single_prefix:
      file_prefix = self.file_prefix
    elif use_prefix:
      file_prefix = dt.now().strftime("%Y%m%d_%H%M%S_")

    save_folder = self.get_models_folder()
    if not to_models:
      save_folder = self.get_output_folder()

    if subfolder_path is not None:
      save_folder = os.path.join(save_folder, subfolder_path.lstrip('/'))

    if not os.path.exists(save_folder):
      os.makedirs(save_folder)

    file_name = os.path.join(save_folder, file_prefix + label)
    if file_name[-len(MDL_EXT):] == MDL_EXT:
      file_name = file_name[:-len(MDL_EXT)]
    if DEBUG:
      self.verbose_log("Saving [...{}]".format(file_name[-40:]))

    save_file_name1 = None
    save_file_name2 = None
    config_file = None

    if save_weights:
      self.P("  Saving model weights...")
      self.P("    WARNING: Model weights and model config saved separately")
      save_file_name1 = file_name + "_weights" + WGH_EXT
      config_file = os.path.join(save_folder, label) + "_mdl_cfg.pkl"
      model.save_weights(save_file_name1)
      dct_cfg = model.get_config()
      with open(config_file, "wb") as f:
        pickle.dump(dct_cfg, f)
      self.P("    Saved: {}".format(save_file_name1))
      self.P("    Saved: {}".format(config_file))

    if save_model:
      self.P("  Saving model...")
      save_file_name2 = file_name + MDL_EXT
      model.save(save_file_name2, include_optimizer=include_optimizer)
      self.P("    Saved: {}".format(save_file_name2))

    self._save_model_config(label, cfg)

    if delete_previous:
      record_previous = True
      for last_saved in self._model_save_last:
        self._model_save_history.append(last_saved)
      self._delete_model_history()

    if record_previous:
      self._model_save_last = []
      for fn in [save_file_name1, save_file_name2, config_file]:
        if fn:
          self._model_save_last.append(fn)

    return file_name + MDL_EXT

  def plot_keras_history(self, keras_history_object,
                         styles=['b-', 'r-'], model_name="",
                         skip_first_epochs=0,
                         plot_pairs=True,
                         plot_all=True,
                         show=False,
                         epochs_thr=4,
                         plot_title='Training history',
                         include_prefix=True,
                         use_single_prefix=True):
    """
    skip_fist_epochs :  skips first history steps  - default 1


    PyPlot style "manual":
        '-' 	solid line style
        '--' 	dashed line style
        '-.' 	dash-dot line style
        ':' 	dotted line style
        '.' 	point marker
        ',' 	pixel marker
        'o' 	circle marker
        'v' 	triangle_down marker
        '^' 	triangle_up marker
        '<' 	triangle_left marker
        '>' 	triangle_right marker
        '1' 	tri_down marker
        '2' 	tri_up marker
        '3' 	tri_left marker
        '4' 	tri_right marker
        's' 	square marker
        'p' 	pentagon marker
        '*' 	star marker
        'h' 	hexagon1 marker
        'H' 	hexagon2 marker
        '+' 	plus marker
        'x' 	x marker
        'D' 	diamond marker
        'd' 	thin_diamond marker
        '|' 	vline marker
        '_' 	hline marker

        The following color abbreviations are supported:
        character 	color
        ‘b’ 	blue
        ‘g’ 	green
        ‘r’ 	red
        ‘c’ 	cyan
        ‘m’ 	magenta
        ‘y’ 	yellow
        ‘k’ 	black
        ‘w’ 	white
    """
    import matplotlib.pyplot as plt
    plt.style.use('ggplot')
    if styles is None:
      styles = ['b-', 'r-']

    keys_lists = [
      ['lr'],
    ]

    plots = []

    if type(keras_history_object) is dict:
      hist = keras_history_object
    else:
      hist = keras_history_object.history

    metrics = list(set(['_'.join(x.split('_')[1:]) for x in hist]))
    ds_keys = list(set(x.split('_')[0] for x in hist))

    for m in metrics:
      keys_lists.append([ds + '_' + m for ds in ds_keys])

    if len(list(hist.values())[0]) < epochs_thr:
      self.P("Cannot show training history with less than {} epochs".format(epochs_thr))
      return
    if plot_pairs:
      found_keys = []
      for keys in keys_lists:
        vals_dict = {}
        for k in keys:
          if k in hist.keys():
            vals_dict[k] = hist[k]
            found_keys.append(k)
        if vals_dict != {}:
          plots.append(vals_dict)
      if plot_all:
        for metric in hist:
          if metric not in found_keys:
            plots.append({metric: hist[metric]})
    else:
      for metric in hist:
        plots.append({metric: hist[metric]})

    nr_plots = len(plots)
    epochs = min([len(hist[x]) for x in hist])  # len(hist[list(hist.keys())[0]])
    nr_sample = min(10, epochs)
    sample_epochs = [0] + np.linspace(1, epochs - 1, nr_sample, dtype=int).tolist()
    fig, ax = plt.subplots(nr_plots, 1, figsize=(20, 5 * nr_plots), sharex=True)
    if nr_plots == 1: ax = [ax]
    max_label = 7
    min_epoch = 1
    max_epoch = epochs + 1
    self.P("Model history results:")
    for pidx, plot in enumerate(plots):
      s_plot = list(plot.keys())
      plot_name = model_name + " "
      short_name = "{}_".format(model_name)
      for i in range(len(s_plot) - 1):
        plot_name += "{} vs. ".format(s_plot[i])
        short_name += "{}VS".format(s_plot[i])
      plot_name += s_plot[-1]
      short_name += s_plot[-1]
      self.P(" Plotting '{}' for {} epochs...".format(plot_name, epochs))
      max_len = max_label
      fmt = "  {:<" + str(max_len + 1) + "} {}"
      s_vals = ""
      for j in sample_epochs:
        s_vals += "{:>5}  ".format(j)
      k = 'Epoch'
      self.P(fmt.format(k + ':', s_vals))
      for i, k in enumerate(plot.keys()):
        vals = plot[k]
        s_vals = ""
        if 'float' in str(type(vals[0])):  # isinstance fails and type fails even worse
          for j in sample_epochs:
            s_vals += "{:>5.3f}  ".format(vals[j])
        else:
          for j in sample_epochs:
            s_vals += "{:>5}  ".format(vals[j])
        self.P(fmt.format(k[:max_label] + ':', s_vals))
        vv = vals[skip_first_epochs:]
        if len(vv) > epochs:
          start_epoch = 0
          min_epoch = 0
          end_epoch = len(vv)
        else:
          start_epoch = 1
          end_epoch = len(vv) + 1
        ax[pidx].plot(np.arange(start_epoch, end_epoch), vv,
                      styles[i], label=k)
      ax[pidx].legend()
      if nr_plots > 1:
        ax[pidx].set_title(plot_name)
    ax[pidx].set_xticks(np.arange(min_epoch, max_epoch))
    self.add_copyright_to_plot(ax[-1])
    fig.tight_layout()
    fig.subplots_adjust(top=0.93)
    s = " " + plot_name if nr_plots == 1 else ""
    fig.suptitle(plot_title + s, fontsize=28)
    self.output_pyplot_image(plt=plt, label=model_name + "_TRAIN_HIST",
                             include_prefix=include_prefix,
                             use_single_prefix=use_single_prefix)
    if show:
      plt.show()  # after show a new figure is created
      self.P("Model history saved and figure shown.")
    else:
      plt.close(fig)
      self.P("Model history saved and figure closed.")
    return



"""
TODO: 
   - merge with MQTT controller implementation - then move to core
   - here just a simple inherited class with init and _CONFIG should reside
  
"""
from time import sleep

from naeural_core.comm import MQTTWrapper
from naeural_core import constants as ct
from naeural_core.comm.base import BaseCommThread

_CONFIG = {
  **BaseCommThread.CONFIG,

  "CONNECTION_FAIL_SLEEP_TIME": 10,

  'VALIDATION_RULES': {
    **BaseCommThread.CONFIG['VALIDATION_RULES'],
  },

  "QOS": 2
}


class MQTTCommThread(BaseCommThread):
  CONFIG = _CONFIG

  def __init__(self, **kwargs):
    self._controller = None
    super(MQTTCommThread, self).__init__(**kwargs)
    return

  def startup(self):
    super().startup()
    return

  @property
  def send_to(self):
    return self._send_to

  @send_to.setter
  def send_to(self, x):
    self._send_to = x
    self._controller.send_channel_name = (self._send_channel_name, self._send_to)
    return

  @property
  def connection(self):
    return self._controller.connection

  def _get_errors(self):
    dct_errs = self._controller.get_connection_issues()
    err_msgs, err_times = [], []
    for err_time in dct_errs:
      err_msgs.append("Disconnected: " + dct_errs[err_time])
      err_times.append(err_time)
    return err_msgs, err_times

  def _get_server_address(self):
    return "{}:{}".format(self._controller.cfg_host, self._controller.cfg_port)

  # base comm abstract implementations
  def _init(self):
    post_on_message = None
    if hasattr(self, '_post_on_message') and self._comm_type == ct.COMMS.COMMUNICATION_DEFAULT:
      # only if type is default we add "on_message" post-processor
      self.P('Using defined `_post_on_message` for incoming commands')
      post_on_message = self._post_on_message

    self._controller = MQTTWrapper(
      log=self.log, config=self._config, recv_buff=self._recv_buff,
      send_channel_name=self._send_channel_name, recv_channel_name=self._recv_channel_name,
      comm_type=self._comm_type,
      post_default_on_message=post_on_message,
      debug_errors=self.cfg_debug_comm_errors,
      connection_name=self.cfg_ee_id,
    )
    self._maybe_reconnect_to_server()
    return

  def _maybe_reconnect_to_server(self):
    if self.connection is None or not self.has_server_conn:
      self.has_send_conn = False
      self.has_recv_conn = False
      dct_ret = self._controller.server_connect()
      self.has_server_conn = dct_ret['has_connection']
      msg = dct_ret['msg']
      msg_type = dct_ret['msg_type']
      if not self.has_server_conn:
        self._nr_conn_retry_iters += 1
        self._total_conn_fails += 1
        self._error_messages.append((msg_type, msg))
        self._error_times.append(self.log.time_to_str())
        ##
        self.P("Multiple connection retry fails occured. Sleeping comms thread for {} sec".format(
          self.cfg_connection_fail_sleep_time), color='r'
        )
        sleep(self.cfg_connection_fail_sleep_time)
        ##
      else:
        self._nr_conn_retry_iters = 0
      self._create_notification(
        notif=msg_type,
        msg=msg
      )
      # self.P(msg)
    # endif
    return

  def _maybe_reconnect_send(self):
    self._maybe_reconnect_to_server()
    if self.has_server_conn:
      self.has_send_conn = True
    else:
      self.has_send_conn = False
    return

  def _maybe_reconnect_recv(self):
    self._maybe_reconnect_to_server()
    if self.has_server_conn and not self.has_recv_conn:
      dct_ret = self._controller.subscribe()
      self.has_recv_conn = dct_ret['has_connection']
      msg = dct_ret['msg']
      msg_type = dct_ret['msg_type']
      if not self.has_recv_conn:
        self._total_conn_fails += 1
        self._error_messages.append((msg_type, msg))
        self._error_times.append(self.log.time_to_str())

      self._create_notification(
        notif=msg_type,
        msg=msg
      )

      topic_name = None
      if self._controller.get_recv_channel_def() is not None:
        topic_name = self._controller.get_recv_channel_def()['TOPIC']

      if self.has_recv_conn:
        self.P("Subscribing succeeded to topic '{}'".format(topic_name), color='g')
      else:
        self.P("Subscribing failed to topic '{}'".format(topic_name), color='r')
    # endif
    return

  def _maybe_fill_recv_buffer(self):
    self._controller.receive()
    return

  def _send(self, data, send_to=None):
    self._controller.send(data, send_to=send_to)
    return

  def _release(self):
    dct_ret = self._controller.release()
    for msg in dct_ret['msgs']:
      self.P(msg)

    self.has_send_conn = False
    self.has_recv_conn = False
    self.has_send_conn = False

    del self._controller
    return



"""
TODO:
  - must find locking methods for free position searching
  - change additional to np.int64 to allow for PID of child process instead of 1 when marking a position as occupied
  - use pid in read_buffer_data to check if the data is requested by the right process
"""

import numpy as np
import sys
import traceback

from naeural_core import DecentrAIObject

__VER__ = '0.2.1.3'
MAXLEN = 50
SHAPE_OFFSET = 10  # on how many elements the shape will be stored


class NumpySharedMemory(DecentrAIObject):
  def __init__(
      self, mem_name, np_shape, np_type,
      create=False, is_buffer=False,
      shape_offset=SHAPE_OFFSET,
      maxlen=MAXLEN, **kwargs
  ):
    """
    Opens an already existing kernel level shared memory depending of python version.
    In case this is used as a buffer there will be an additional memory zone used
    to store the availability of the memory and the shape of the numpy array.
    It will have the shape (maxlen, 1 + shape_offset) and each line will be structured as follows:
    - the first element will dictate if the line is available or not
    - the next elements will store the shape of the numpy array
  
    Parameters
    ----------
    mem_name : str
      shared memory name as created by its originator.

    np_shape : tuple
      shape of the target numpy ndarray.
      
    np_type : str
      type of the target numpy ndarray.

    create : bool
      if True, the shared memory is created, otherwise it is opened.

    is_buffer : bool
      if True, the shared memory is treated as a buffer, otherwise it is treated as a single zone of memory.

    shape_offset : int
      Will be used when the shared memory is treated as a buffer.
      This will dictate how many elements will be used to store the shape of the numpy array.

    maxlen : int
      Will be used then the shared memory is treated as a buffer.
      
    log : Logger
      parent instance of the Logger
      
  
    Returns
    -------
    Nothing.  
    """
    self.initiator = create
    self.error_message = None
    self.initialized = False
    self.version = __VER__
    self._mem_name = mem_name
    self._np_shape = np_shape
    self._np_type = np_type
    prod_shape = np.prod(np_shape, dtype='uint64')
    self._np_bytes = prod_shape * np.dtype(np_type).itemsize
    self.is_buffer = is_buffer
    self.maxlen = maxlen
    self.shape_offset = shape_offset
    self.buffer_position = 0
    if self.is_buffer:
      self._np_shape = (maxlen, prod_shape)
      self._np_bytes = maxlen * self._np_bytes
      self._max_element_size = prod_shape
    self._mmap_obj = None
    self._shm = None
    super(NumpySharedMemory, self).__init__(**kwargs)
    return

  @property
  def uses_mp(self):
    return self.log.python_minor >= 8

  @property
  def shm_type(self):
    if self.uses_mp:
      return "multiprocessing.SharedMemory"
    else:
      return "posix_ipc"      

  def get_instance_type(self, initiator=False):
    return "Creator" if initiator else "User"

  @property
  def instance_type(self):
    return self.get_instance_type(self.initiator)

  def _set_error(self, err):
    self.error_message = err
    self.P(err, color='error')
    self.initialized = False
    return

  def _add_warning(self, warn):
    self.P(warn, color='warning')
    return

  def __repr__(self):
    return f"NumpySharedMemory({self._mem_name}, {self._np_bytes}, {self._np_shape}, {self._np_type})"
  
  def startup(self):
    super().startup()
    self.error_message = self._init_mem()
    if self.error_message is None:
      self.initialized = True
      self.P(f"Shared memory '{self._mem_name}'[{self.instance_type}] successfully initialized")
    # endif initialized
    return
  
  
  def _init_mem(self):
    self.P("Initializing shmem '{}' using {}".format(self._mem_name, self.shm_type))
    if self.uses_mp:
      msg = self._mp_open()
    else:
      msg = self._posix_ipc_open()
    return msg
      
  
  def P(self, s, **kwargs):
    super().P(s, prefix=True, **kwargs)
    return

  def _posix_ipc_open(self):
    try:
      import posix_ipc 
      import mmap

      def initialize_posix_shared_memory(
          flags, mode, mem_name, mem_size, np_shape, np_type
      ):
        self.P("Initializing ({} - {}) posix_ipc SharedMemory {}".format(
          flags, mode, mem_name))

        shm = posix_ipc.SharedMemory(
          name=mem_name,
          # posix_ipc.O_RDWR is not available in 1.0.4 version O_RDWR=2 O_CREAT=512, CRRW=514
          flags=flags,
          size=int(mem_size),
        )
        # mmap
        self.P(f"Initializing mmap on fd: {shm.fd}, size: {shm.size}")
        mmap_obj = mmap.mmap(shm.fd, shm.size)
        # memoryview of mmap
        mmap_ptr = memoryview(mmap_obj)
        # after mmap we are done with the fd
        shm.close_fd()
        # finally the numpy
        self.P("Initializing shmem ndarray")
        np_mem = np.ndarray(
          shape=np_shape,
          dtype=np_type,
          buffer=mmap_ptr,
        )
        self.P("Done initializing shmem.")

        return shm, mmap_obj, np_mem

      # open shm
      _O_RDWR = 2
      _O_CREAT = 64
      flags = _O_RDWR if not self.initiator else _O_CREAT | _O_RDWR
      mode = "creating" if self.initiator else "open r/w"
      mem_init_kwargs = {
        'flags': flags,
        'mode': mode,
        'mem_name': self._mem_name,
        'mem_size': self._np_bytes,
        'np_shape': self._np_shape,
        'np_type': self._np_type,
      }
      self._shm, self._mmap_obj, self._np_mem = initialize_posix_shared_memory(**mem_init_kwargs)
      if self.is_buffer:
        # additional memory for buffer mode
        # This will be used to store the availability of memory and the shape of the numpy array
        mem_init_kwargs['mem_name'] = f'{self._mem_name}_additional'
        mem_init_kwargs['np_shape'] = (self.maxlen, 1 + self.shape_offset)
        mem_init_kwargs['np_type'] = np.int32
        mem_init_kwargs['mem_size'] = self.maxlen * (1 + self.shape_offset) * np.dtype(np.int32).itemsize
        self._shm_additional, self._mmap_obj_additional, self._np_mem_additional = initialize_posix_shared_memory(**mem_init_kwargs)
      # endif buffer mode
      error_msg = None
    except:
      full_err = traceback.format_exc()
      error_msg =  "ERROR in posix_ipc ShareMemory('{}') open: {}\n{}".format(
        self._mem_name,
        sys.exc_info()[0], full_err,
      )
      self._set_error(error_msg)
      self._create_notification(
         notif='EXCEPTION',
         msg=error_msg,
      )
      # fake numpy buffer
      self._np_mem = np.zeros(
        shape=self._np_shape,
        dtype=self._np_type,
      )
      if self.is_buffer:
        self._np_mem_additional = np.zeros(
          shape=(self.maxlen, 1 + self.shape_offset),
          dtype=np.int32,
        )
    return error_msg
  
  def _mp_open(self):
    try:
      from multiprocessing import shared_memory

      def initialize_mp_shared_memory(
          create, mem_name, mem_size, np_shape, np_type
      ):
        try:
          # Attempt to create the shared memory object.
          shm = shared_memory.SharedMemory(
            create=create,
            name=mem_name,
            size=int(mem_size),
          )
        except Exception as exc:
          # If the shared memory object already exists, attempt to open it.
          # If the shared memory object does not exist, create it.
          initial_instance_type = self.get_instance_type(create)
          current_instance_type = self.get_instance_type(not create)
          self.P(f"Failed to create shared memory '{mem_name}'[{initial_instance_type}]: {exc}\n"
                 f"Attempting to open shared memory '{mem_name}'[{current_instance_type}]")
          create = not create
          shm = shared_memory.SharedMemory(
            create=create,
            name=mem_name,
            size=int(mem_size),
          )
        # endtry create shared memory
        # Create a numpy array from the shared memory object.
        # The numpy array will be used to read and write data to the shared memory object.
        np_mem = np.ndarray(
          np_shape,
          dtype=np_type,
          buffer=shm.buf,
        )
        return shm, np_mem, create

      mem_init_kwargs = {
        'create': self.initiator,
        'mem_name': self._mem_name,
        'mem_size': self._np_bytes,
        'np_shape': self._np_shape,
        'np_type': self._np_type,
      }
      self._shm, self._np_mem, self.initiator = initialize_mp_shared_memory(**mem_init_kwargs)
      if self.is_buffer:
        # additional memory for buffer mode
        # This will be used to store the availability of memory and the shape of the numpy array
        mem_init_kwargs['mem_name'] = f'{self._mem_name}_additional'
        mem_init_kwargs['np_shape'] = (self.maxlen, 1 + self.shape_offset)
        mem_init_kwargs['np_type'] = np.int32
        mem_init_kwargs['mem_size'] = int(self.maxlen * (1 + self.shape_offset) * np.dtype(np.int32).itemsize)
        mem_init_kwargs['create'] = self.initiator
        self._shm_additional, self._np_mem_additional, _ = initialize_mp_shared_memory(**mem_init_kwargs)
      # endif buffer mode
      error_msg = None
    except:
      full_err = traceback.format_exc()
      error_msg =  "ERROR in multiprocessing ShareMemory('{}') open: {}\n{}".format(
        self._mem_name,
        sys.exc_info()[0], full_err,
      )
      self._set_error(error_msg)
      self._create_notification(
         notif='EXCEPTION',
         msg=error_msg,
      )
      # fake numpy buffer
      self._np_mem = np.zeros(
        shape=self._np_shape,
        dtype=self._np_type,        
      )
      if self.is_buffer:
        self._np_mem_additional = np.zeros(
          shape=(self.maxlen, 1 + self.shape_offset),
          dtype=np.int32,
        )
      # endif buffer mode
      
    return error_msg

  def get_position(self, position):
    """
    Returns a valid position of the shared memory buffer.
    Parameters
    ----------
    position : int or list or tuple or None

    Returns
    -------
    list - array of integers with the position
    """
    if position is None:
      return []
    if isinstance(position, (int, np.integer)):
      position = [position]
    if len(position) > len(self._np_shape):
      position = position[:len(self._np_shape)]
    nr_pos = min(len(position), len(self._np_shape))
    for i in range(nr_pos):
      if position[i] >= self._np_shape[i]:
        position[i] = self._np_shape[i] - 1
    return position

  def read(self, position=None):
    pos = self.get_position(position)
    if self.initialized:
      return self._np_mem[tuple(pos)].copy()
    return None
  
  def write(self, val, position=None):
    if not self.initialized:
      return False
    pos = self.get_position(position)
    curr_mem = self._np_mem[tuple(pos)]
    if isinstance(val, np.ndarray):
      if curr_mem.shape != val.shape:
        section_str = f" [{pos}]" if len(pos) > 0 else ""
        err_msg = f"Shape mismatch for input {val.shape} vs numpy shmem{section_str} {curr_mem.shape}"
        self._set_error(err_msg)
        return False
      curr_mem[:] = val[:]
    elif isinstance(val, (int, float, np.integer)):
      curr_mem[:] = val
    else:
      self._set_error("Uknown data type {} for NumpySharedMemory".format(
        type(val)
      ))
      return False
    return True

  def first_buffer_position(self):
    if not self.initialized:
      return None
    if self.is_buffer:
      for i in range(self.maxlen):
        if self._np_mem_additional[i, 0] == 0:
          self._np_mem_additional[i, 0] = 1
          return i
    return None

  def check_data(self, data):
    """
    Checks if the data is a valid for the shared memory buffer.
    This is used only in buffer mode.
    Parameters
    ----------
    data : any

    Returns
    -------
    True if the data is valid, False otherwise.
    """
    valid = True
    if not isinstance(data, np.ndarray):
      self._add_warning("Data is not a numpy array")
      valid = False
    if np.prod(data.shape) > self._max_element_size:
      self._add_warning("Data of shape {} is larger than the maximum size {}".format(
        data.shape, self._max_element_size
      ))
      valid = False
    if len(data.shape) > self.shape_offset:
      self._add_warning("Data shape {} has more than {} dimensions".format(
        data.shape, self.shape_offset
      ))
      valid = False
    return valid

  def add_data_to_buffer(self, data: np.ndarray):
    """
    Adds data to the buffer memory zone.
    Parameters
    ----------
    data : np.ndarray of maximum self.shape_offset dimensions

    Returns
    -------
    res : int or None - the position in the buffer or None if:
     - the buffer is full
     - the buffer is not initialized
     - the data is not valid
    """
    if not self.check_data(data):
      return None
    insert_pos = self.first_buffer_position()
    if insert_pos is not None:
      self._np_mem_additional[insert_pos, 1:1 + len(data.shape)][:] = data.shape
      reshaped_data = data.reshape(-1)
      self._np_mem[insert_pos, :len(reshaped_data)][:] = reshaped_data[:]
      return insert_pos
    return None

  def read_buffer_data(self, position, pid=None):
    """
    Reads data from the buffer memory zone.
    Parameters
    ----------
    position : int - the position in the buffer
    pid : int - the process id of the process requesting the data

    Returns
    -------
    data : np.ndarray or None - the data from the buffer or None if:
      - the buffer is not initialized
      - the position is invalid
      - the data is not available
    """
    position = int(position)
    if -1 < position < self.maxlen and self.initialized:
      if self._np_mem_additional[position, 0] == 1:
        shape = self._np_mem_additional[position, 1:1 + self.shape_offset]
        shape = shape[shape > 0]
        data = self._np_mem[position, :np.prod(shape)].reshape(shape)
        self._np_mem_additional[position, :] = 0
        return data
      # endif data is available
    # endif position is valid and memory initialized
    return None

  def shutdown(self):
    self.P(f"Shutdown NumpySharedMemory {self._mem_name}[{self.instance_type}]...")
    if self._mmap_obj is not None:
      self._mmap_obj.close()
      if self.is_buffer:
        self._mmap_obj_additional.close()
    if self._shm is not None:
      del self._np_mem
      self._shm.close()
      if self.initiator:
        self._shm.unlink()
      if self.is_buffer:
        del self._np_mem_additional
        self._shm_additional.close()
        if self.initiator:
          self._shm_additional.unlink()
    return


"""SHM CLASS END"""


class NumpySharedMemoryPlaceholder:
  def __init__(self, shm_idx):
    self.__shm_idx = shm_idx
    return

  def get_shm_idx(self):
    return self.__shm_idx


def ndarray_to_shm(data: np.ndarray, shm: NumpySharedMemory, debug: bool = False) -> NumpySharedMemoryPlaceholder:
  """
  Helper function to replace a numpy ndarray with a shared memory object.
  Parameters
  ----------
  data - np.ndarray - the numpy array to be replaced
  shm - NumpySharedMemory - the shared memory object to be used
  debug - bool - if True, the function will print debug information

  Returns
  -------
  res - NumpySharedMemoryPlaceholder - the placeholder object
  """
  if debug:
    shm.log.P(f"Writing data of shape {data.shape if isinstance(data, np.ndarray) else None} to shared memory.")
  idx = shm.add_data_to_buffer(data)
  if debug:
    shm.log.P(f"Data written to shared memory index {idx}")
  return NumpySharedMemoryPlaceholder(idx)


def replace_ndarray_with_shm(data: any, shm: NumpySharedMemory, debug: bool = False) -> any:
  """
  Replaces all numpy ndarrays from data with shared memory placeholders.
  Parameters
  ----------
  data - any - the data to be processed
  shm - NumpySharedMemory or None - the shared memory object to be used or None if no replacement is needed
  debug - bool - if True, the function will print debug information

  Returns
  -------
  res - any - the data with all numpy ndarrays replaced with shared memory placeholders or left unchanged
  """
  if shm is None:
    return data
  if isinstance(data, list):
    res = [None] * len(data)
    for i in range(len(data)):
      if isinstance(data[i], np.ndarray):
        res[i] = ndarray_to_shm(data[i], shm, debug)
      else:
        res[i] = replace_ndarray_with_shm(data[i], shm, debug)
    # endfor data
  elif isinstance(data, dict):
    res = {}
    for k in data:
      if isinstance(data[k], np.ndarray):
        res[k] = ndarray_to_shm(data[k], shm, debug)
      else:
        res[k] = replace_ndarray_with_shm(data[k], shm, debug)
    # endfor data
  elif isinstance(data, np.ndarray):
    res = ndarray_to_shm(data, shm, debug)
  else:
    res = data
  return res


def shm_to_ndarray(data: NumpySharedMemoryPlaceholder, shm: NumpySharedMemory, debug: bool = False) -> np.ndarray:
  """
  Helper function to replace a shared memory object with a numpy ndarray.
  Parameters
  ----------
  data - NumpySharedMemoryPlaceholder - the placeholder object to be replaced
  shm - NumpySharedMemory - the shared memory object to be used
  debug - bool - if True, the function will print debug information

  Returns
  -------
  res - np.ndarray - the numpy array
  """
  shm_idx = data.get_shm_idx()
  if debug:
    shm.log.P(f"Reading data from shared memory index {shm_idx}")
  res = shm.read_buffer_data(shm_idx) if shm_idx is not None else None
  if debug:
    shm.log.P(f"Read data of shape {res.shape if isinstance(res, np.ndarray) else None} from shared memory index {shm_idx}.")
  return res


def replace_shm_with_ndarray(data: any, shm: NumpySharedMemory, debug: bool = False) -> any:
  """
  Replaces all shared memory placeholders from data with numpy ndarrays.
  Parameters
  ----------
  data - any - the data to be processed
  shm - NumpySharedMemory or None - the shared memory object to be used or None if no replacement is needed
  debug - bool - if True, the function will print debug information

  Returns
  -------
  res - any - the data with all shared memory placeholders replaced with numpy ndarrays or left unchanged
  """
  if shm is None:
    return data
  if isinstance(data, list):
    res = [None] * len(data)
    for i in range(len(data)):
      if isinstance(data[i], NumpySharedMemoryPlaceholder):
        res[i] = shm_to_ndarray(data[i], shm, debug)
      else:
        res[i] = replace_shm_with_ndarray(data[i], shm, debug)
    # endfor data
  elif isinstance(data, dict):
    res = {}
    for k in data:
      if isinstance(data[k], NumpySharedMemoryPlaceholder):
        res[k] = shm_to_ndarray(data[k], shm, debug)
      else:
        res[k] = replace_shm_with_ndarray(data[k], shm, debug)
    # endfor data
  elif isinstance(data, NumpySharedMemoryPlaceholder):
    res = shm_to_ndarray(data, shm, debug)
  else:
    res = data
  return res


if __name__ == '__main__':
  RECREATE = False

  from time import sleep
  def test_shm_release(release_method='del'):
    from time import time
    from multiprocessing import shared_memory
    curr_ts = time()
    shm_kwargs = {
      'name': f'test_shm_{str(curr_ts)}',
      'size': 100,
    }
    shm1 = shared_memory.SharedMemory(**shm_kwargs, create=True)
    print(f'{shm1} created')

    kwargs2 = {**shm_kwargs, 'size': 0}
    shm2 = shared_memory.SharedMemory(**kwargs2, create=False)
    print(f'{shm2} created')

    print('Attempting to release the memory...')
    if release_method == 'del':
      del shm1
    elif release_method == 'unlink':
      shm1.unlink()
      shm1.close()
    elif release_method == 'close':
      shm1.close()
    elif release_method == 'buffrelease':
      shm1.buf.release()
    else:
      print(f'Unknown release method {release_method}')
      return

    print(f'Shared memory deleted using {release_method}')
    # print('Attempting freeing the memory...')

    if RECREATE:
      try:
        shm3 = shared_memory.SharedMemory(**shm_kwargs, create=True)
        print(f'{shm3} created')
      except:
        print('Failed to recreate the memory')
    else:
      shm2.close()
      # if release_method == 'del':
      #   del shm2
      # elif release_method == 'unlink':
      #   shm2.unlink()
      # elif release_method == 'close':
      #   shm2.close()
      # elif release_method == 'buffrelease':
      #   shm2.buf.release()
      # else:
      #   print(f'Unknown release method {release_method}')
      #   return
      print('Shared memory reference deleted')
      print('Attempting recreating the memory...')
      try:
        shm3 = shared_memory.SharedMemory(**shm_kwargs, create=True)
        print(f'{shm3} created')
        del shm3
      except Exception as exc:
        print(f'Failed to recreate the memory:\n\t{exc}')
        sleep(30)
    return

  def full_test_shm_release():
    for close_method in ['del', 'unlink', 'close', 'buffrelease']:
      print(f'###Starting test for {close_method}...####')
      test_shm_release(close_method)
      print(f'###Test for {close_method} done###########')
    while True:
      sleep(1000)
    return

  full_test_shm_release()
  exit(0)

  from naeural_core import Logger
  l = Logger('SHMT', base_folder='.', app_folder='_cache')
  mem_name = 'test1'
  shm1 = NumpySharedMemory(
    mem_name=mem_name, 
    mem_size=100, 
    np_shape=(10,10), 
    np_type=np.uint8, 
    log=l,
    create=True,
  )
  
  if shm1.initialized:
    shm1.write(2)
    
    shm2 = NumpySharedMemory(
      mem_name=mem_name, 
      mem_size=100, 
      np_shape=(10,10), 
      np_type=np.uint8, 
      log=l,
      create=False,
    )
    
    res = shm2.read()
    
    print(res)
    
    shm1.shutdown()
    shm2.shutdown()
    
  
  

    
    


"""
{
  "NAME" : "peer_config_pipeline",
  "TYPE" : "NetworkListener",
  
  "PATH_FILTER" : [
      null, null, 
      ["UPDATE_MONITOR_01", "NET_MON_01"],
      null
    ],
  "MESSAGE_FILTER" : {},
  
  "PLUGINS" : [
    {
      "SIGNATURE" : "NET_CONFIG_MONITOR",
      "INSTANCES" : [
        {
          "INSTANCE_ID" : "DEFAULT"
        }
      ]
    }
  ]
}


"""
from naeural_core.business.base.network_processor import NetworkProcessorPlugin 


__VER__ = '1.1.0'

_CONFIG = {
  
  **NetworkProcessorPlugin.CONFIG,
  
  'ALLOW_EMPTY_INPUTS' : True,
  
  'PLUGIN_LOOP_RESOLUTION' : 50, # we force this to be 50 Hz from the standard 20 Hz  
  'MAX_INPUTS_QUEUE_SIZE' : 128, # increase the queue size to 128 from std 1
  

  # DEBUGGING
  "FULL_DEBUG_PAYLOADS" : False,     
  "VERBOSE_NETCONFIG_LOGS" : False, 
  # END DEBUGGING
  
  'PROCESS_DELAY' : 1,
  
  # each cfg_send_get_config_each seconds we will send requests to the nodes that allow 
  # us to get their pipelines and that have not been requested in the last 
  # cfg_node_request_configs_each seconds.
  'SEND_GET_CONFIG_EACH' : 600, # runs the send request logic every 10 minutes
  'NODE_REQUEST_CONFIGS_EACH' : 1200, # minimum time between requests to the same node
  
  # sent each cfg_send_to_allowed_each seconds the pipeline configuration to the allowed nodes
  'SEND_TO_ALLOWED_EACH' : 600, # runs the send to allowed nodes logic every 10 minutes
  
  
  'SHOW_EACH' : 60,
  
  'DEBUG_NETMON_COUNT' : 1,
  
  'VALIDATION_RULES' : {
    **NetworkProcessorPlugin.CONFIG['VALIDATION_RULES'],
  },
}

class NetConfigMonitorPlugin(NetworkProcessorPlugin):
  CONFIG = _CONFIG
  
  CT_PIPELINE = "PIPELINES"
  CT_PLG_STATUSES = "PLUGIN_STATUSES"
  
  
  def Pd(self, *args, **kwargs):
    """
    This function will print the debug messages.
    """
    if self.cfg_verbose_netconfig_logs:
      self.P(*args, **kwargs)
    return
  
  
  def on_init(self):   
    self.P("Network fleet peer configuration monitor initializing...")
    self.__last_data_time = 0
    self.__new_nodes_this_iter = 0
    self.__last_shown = 0
    self.__recvs = self.defaultdict(int)
    self.__initial_send = False
    self.__last_pipelines = None
    self.__allowed_nodes = {} # contains addresses with no prefixes
    self.__last_sent_to_allowed = 0
    self.__debug_netmon_count = self.cfg_debug_netmon_count
    self._get_active_plugins_instances = self.global_shmem.get("get_active_plugins_instances")
    if not callable(self._get_active_plugins_instances):
      self.P(" ERROR: `get_active_plugins_instances` not found!", color='r', boxed=True)
    return
  
  
  def __check_dct_metadata(self):
    """ This is a debug function that checks the metadata of the dataapi stream. """
    stream_metadata = self.dataapi_stream_metadata()
    if stream_metadata is not None:
      if self.cfg_verbose_netconfig_logs:
        self.P(f"Stream metadata:\n {self.json_dumps(stream_metadata, indent=2)}")
    return
  
  
  def __update_allowed_nodes(self, addr, pipelines):
    sender_no_prefix = self.bc.maybe_remove_prefix(addr)
    if sender_no_prefix not in self.__allowed_nodes:
      self.__allowed_nodes[sender_no_prefix] = {}
    #endif sender_no_prefix not in __allowed_nodes
    self.__allowed_nodes[sender_no_prefix]["pipelines"] = pipelines
    self.__allowed_nodes[sender_no_prefix]["last_config_get"] = self.time()
    self.__allowed_nodes[sender_no_prefix]["is_online"] = True
    return


  def __preprocess_current_network_data(self, current_network):
    for _, v in current_network.items():
      addr = v.get(self.const.PAYLOAD_DATA.NETMON_ADDRESS)
      if addr is None:
        continue
      v[self.const.PAYLOAD_DATA.NETMON_ADDRESS] = self.bc.maybe_remove_addr_prefix(addr)
    return current_network


  def __get_active_nodes(self, netmon_current_network : dict) -> dict:
    """
    Returns a dictionary with the active nodes in the network.
    """
    active_network = {
      v[self.const.PAYLOAD_DATA.NETMON_ADDRESS] : v 
      for k, v in netmon_current_network.items() 
      if v.get(self.const.PAYLOAD_DATA.NETMON_STATUS_KEY, False) == self.const.DEVICE_STATUS_ONLINE
    }    
    return active_network


  def __get_active_nodes_summary_with_peers(self, netmon_current_network: dict):
    """
    Looks in all whitelists and finds the nodes that is allowed by most other nodes.
    
    """
    node_coverage = {}
    
    active_network = self.__get_active_nodes(netmon_current_network)
    
    for addr in active_network:
      node_coverage[addr] = 0
    #endfor initialize node_coverage 
    
    whitelists = [x.get("whitelist", []) for x in active_network.values()]
    for whitelist in whitelists:
      for ee_addr in whitelist:
        if ee_addr not in active_network:
          continue # this address is not active in the network so we skip it
        if ee_addr not in node_coverage:
          node_coverage[ee_addr] = 0
        node_coverage[ee_addr] += 1
    coverage_list = [(k, v) for k, v in node_coverage.items()]
    coverage_list = sorted(coverage_list, key=lambda x: x[1], reverse=True)

    result = self.OrderedDict()
    my_addr = self.bc.maybe_remove_prefix(self.ee_addr)
    
    for i, (ee_addr, coverage) in enumerate(coverage_list):
      is_online = active_network.get(ee_addr, {}).get("working", False) == self.const.DEVICE_STATUS_ONLINE
      result[ee_addr] = {
        "peers" : coverage,
        "eeid" : active_network.get(ee_addr, {}).get("eeid", "UNKNOWN"),
        'ver'  : active_network.get(ee_addr, {}).get("version", "UNKNOWN"),
        'is_supervisor' : active_network.get(ee_addr, {}).get("is_supervisor", False),
        'allows_me' : my_addr in active_network.get(ee_addr, {}).get("whitelist", []),
        'online' : is_online,
        'whitelist' : active_network.get(ee_addr, {}).get("whitelist", []),
      }
    return result


  def __maybe_review_known(self):
    """
    This function will show the known nodes every `cfg_show_each` seconds.
    """
    if ((self.time() - self.__last_shown) < self.cfg_show_each):
      return
    self.__last_shown = self.time()
    msg = "Known nodes: "
    if len(self.__allowed_nodes) == 0:
      msg += "\n=== No allowed nodes to show ==="
    else:
      for addr in self.__allowed_nodes:
        eeid = self.netmon.network_node_eeid(addr)
        pipelines = self.__allowed_nodes[addr].get("pipelines", [])
        names = [p.get("NAME", "NONAME") for p in pipelines]
        me_msg = ""
        prefixed_addr = self.bc.maybe_add_prefix(addr)
        if prefixed_addr == self.ee_addr:
          pipelines = self.node_pipelines
          names = [p.get("NAME", "NONAME") for p in pipelines]
          me_msg = " (ME)"
        msg += f"\n  - '{eeid}' <{addr}>{me_msg} has {len(pipelines)} pipelines: {names}"
      #endfor __allowed_nodes    
    self.P(msg)
    return
  
  
  def __send_get_cfg(self, node_addr):
    """
    Sends a request to a node or a list of nodes to get their configuration.
    """
    if isinstance(node_addr, list):
      node_addr = [self.bc.maybe_add_prefix(x) for x in node_addr if self.netmon.network_node_is_online(x)]
      node_ee_id = [self.netmon.network_node_eeid(x) for x in node_addr]
    else:
      node_addr = self.bc.maybe_add_prefix(node_addr) # add prefix if not present otherwise the protocol will fail
      node_ee_id = self.netmon.network_node_eeid(node_addr)

    if self.cfg_verbose_netconfig_logs:
      self.P(f"Sending {self.const.NET_CONFIG.REQUEST_COMMAND} to '{node_ee_id}' <{node_addr}>...")    
    payload = {
      self.const.NET_CONFIG.NET_CONFIG_DATA : {
        self.const.NET_CONFIG.OPERATION   : self.const.NET_CONFIG.REQUEST_COMMAND,
        self.const.NET_CONFIG.DESTINATION : node_addr,
      },
    }
    self.send_encrypted_payload(node_addr=node_addr, **payload)
    return
  
  
  def __send_set_cfg(self, node_addr):
    if isinstance(node_addr, list):
      node_addr = [self.bc.maybe_add_prefix(x) for x in node_addr if self.netmon.network_node_is_online(x)]
      node_ee_id = [self.netmon.network_node_eeid(x) for x in node_addr]
    else:
      node_addr = self.bc.maybe_add_prefix(node_addr) # add prefix if not present otherwise the protocol will fail
      node_ee_id = self.netmon.network_node_eeid(node_addr)

    my_pipelines = self.node_pipelines
    
    if self.cfg_verbose_netconfig_logs:
      self.P(f"Sending {self.const.NET_CONFIG.STORE_COMMAND}:{len(my_pipelines)} to requester '{node_ee_id}' <{node_addr}>...")
      
    statuses = []
    if self._get_active_plugins_instances is not None and callable(self._get_active_plugins_instances):
      statuses = self._get_active_plugins_instances()
    payload = {
      self.const.NET_CONFIG.NET_CONFIG_DATA : {
        self.const.NET_CONFIG.OPERATION : self.const.NET_CONFIG.STORE_COMMAND,
        self.const.NET_CONFIG.DESTINATION : node_addr,
        self.CT_PIPELINE : my_pipelines,
        self.CT_PLG_STATUSES : statuses,
      }
    }
    self.send_encrypted_payload(
      node_addr=node_addr,
      **payload
    )
    return    


  def __maybe_send_requests(self):
    """
    This function will send requests to the nodes that allow us to get their pipelines.
    
    This function will run every `cfg_send_get_config_each` seconds and check if there is any node that
    allow current node to request their pipelines and that node has not been requested 
    in the last `cfg_node_request_configs_each` seconds.
    
    """
    if self.time() - self.__last_data_time > self.cfg_send_get_config_each:
      if len(self.__allowed_nodes) == 0:
        self.Pd("No allowed nodes to send requests to. Waiting for network data...")
        self.__last_data_time = self.time() - self.cfg_send_get_config_each + 10 # we force after 10 seconds to trigger
      else:
        self.__last_data_time = self.time()
        self.Pd(f"Initiating pipeline requests to {len(self.__allowed_nodes)} allowed nodes...")
        to_send = []
        for node_addr in self.__allowed_nodes:
          last_request = self.__allowed_nodes[node_addr].get("last_config_get", 0)
          if (self.time() - last_request) > self.cfg_node_request_configs_each and self.__allowed_nodes[node_addr]["is_online"]:
            to_send.append(node_addr)
          #endif enough time since last request of this node
        #endfor __allowed_nodes
        if len(to_send) == 0:
          self.Pd("No nodes need update.")
        else:
          self.Pd(f"Sending requests to {len(to_send)} nodes...")        
          # now send some requests
          self.__send_get_cfg(node_addr=to_send)
          for node_addr in to_send:
            self.__allowed_nodes[node_addr]["last_config_get"] = self.time()
          #endfor to_send
        #endif len(to_send) == 0
      #endif have allowed nodes
    #endif time to send
    return
  
  
  def __check_allowed_request(self, node_addr):
    allowed_list = self.bc.get_whitelist()
    node_addr = self.bc.maybe_remove_addr_prefix(node_addr)
    result = True
    if node_addr not in allowed_list:
      result = False
    return result
  
  
  def __maybe_send_configuration_to_allowed(self):
    """
    This function will send the configuration to all the allowed nodes when the configuration is updated or when the node starts.
    """
    allowed_list = self.bc.get_whitelist(with_prefix=True)
    must_distribute = False    
    if not self.__initial_send:
      self.P(f"Sending initial configuration to {len(allowed_list)} allowed nodes.")
      must_distribute = True
      self.__initial_send = True
      
    if not must_distribute and self.__last_pipelines != self.node_pipelines:
      self.P("Sending updated configuration to all allowed nodes.")
      must_distribute = True
      self.__last_pipelines = self.deepcopy(self.node_pipelines)
    
    if not must_distribute and (self.time() - self.__last_sent_to_allowed > self.cfg_send_to_allowed_each):
      self.P(f"Sending configuration to all allowed nodes at timeout={self.cfg_send_to_allowed_each}.")
      must_distribute = True      
      
    if must_distribute:
      if len(allowed_list) > 0:
        self.__send_set_cfg(node_addr=allowed_list)
      else:
        self.P("No allowed nodes to send configuration to although we have updated configuration.")
      self.__last_sent_to_allowed = self.time()      
    #endif must_distribute
    return
    
  

  @NetworkProcessorPlugin.payload_handler()
  def default_handler(self, payload: dict):
    sender = payload.get(self.const.PAYLOAD_DATA.EE_SENDER, None)
    receiver = payload.get(self.const.PAYLOAD_DATA.EE_DESTINATION, None)
    if not isinstance(receiver, list):
      receiver = [receiver]
    
    if self.ee_addr not in receiver:
      if self.cfg_verbose_netconfig_logs:
        payload_path = payload.get(self.const.PAYLOAD_DATA.EE_PAYLOAD_PATH, [None, None, None, None])
        is_encrypted = payload.get(self.const.PAYLOAD_DATA.EE_IS_ENCRYPTED, False)
        self.P("Received {}payload for <{}> but I am <{}>: {}'".format(
        "ENCRYPTED " if is_encrypted else "", receiver, self.ee_addr, payload_path), color='r'
        )
      return
    else:
      if self.cfg_verbose_netconfig_logs:
        self.P(f"Received request from <{sender}>")
    
    sender_no_prefix = self.bc.maybe_remove_prefix(sender)
    sender_id = self.netmon.network_node_eeid(sender_no_prefix)
    is_encrypted = payload.get(self.const.PAYLOAD_DATA.EE_IS_ENCRYPTED, False)

    if is_encrypted:
      decrypted_data = self.receive_and_decrypt_payload(data=payload)
      if decrypted_data is not None:
        net_config_data = decrypted_data.get(self.const.NET_CONFIG.NET_CONFIG_DATA, {})
        op = net_config_data.get(self.const.NET_CONFIG.OPERATION, "UNKNOWN")
        # now we can process the data based on the operation
        if op == self.const.NET_CONFIG.STORE_COMMAND:
          received_pipelines = net_config_data.get(self.CT_PIPELINE, [])    
          received_plugins_statuses = net_config_data.get(self.CT_PLG_STATUSES, [])
          if self.cfg_verbose_netconfig_logs:            
            self.P("Received {} data from '{}' <{}>'.\n  - Pipelines: {}\n  - Plugins: {}".format(
              self.const.NET_CONFIG.STORE_COMMAND, sender_id, sender, 
              len(received_pipelines), len(received_plugins_statuses)
            ))
          # process in local cache
          self.__update_allowed_nodes(sender_no_prefix, received_pipelines)
          # now we can add the pipelines to the netmon cache
          self.netmon.register_node_pipelines(
            addr=sender_no_prefix, pipelines=received_pipelines,
            plugins_statuses=received_plugins_statuses,
            verbose=self.cfg_verbose_netconfig_logs
          )
        #finished SET_CONFIG
        
        elif op == self.const.NET_CONFIG.REQUEST_COMMAND:
          self.P(f"Received {self.const.NET_CONFIG.REQUEST_COMMAND} data from '{sender_id}' <{sender}'.")
          ###
          ### At this point we can check if the sender is allowed to request our pipelines
          ### While this is handled naturally by the comms for the normal commands in this
          ### case we need to check the whitelist directly.
          ###
          if self.__check_allowed_request(sender_no_prefix):
            self.__send_set_cfg(sender)
          else:
            self.P(f"Node '{sender_id}' <{sender}> is not allowed to request my pipelines. This behavior should be recorded!", color='r')
        #finished GET_CONFIG
        #endif ops
    else:
      self.P("Received unencrypted data. Dropping.", color='r')
    return  
  
  
  @NetworkProcessorPlugin.payload_handler("NET_MON_01")
  def netmon_handler(self, data : dict):
    """
    This function will process the NET_MON_01 data and update the allowed nodes list.
    The objective is to keep track of the nodes that we are allowed to send requests to.
    """
    current_network = data.get(self.const.PAYLOAD_DATA.NETMON_CURRENT_NETWORK, {})
    if len(current_network) == 0:
      self.P(f"[netmon_handler] Received NET_MON_01 data without {self.const.PAYLOAD_DATA.NETMON_CURRENT_NETWORK}.", color='r ')
    else:
      # here we will remove the prefix from each "address" within the nodes info
      current_network = self.__preprocess_current_network_data(current_network)
      # from this point on we will work with the no-prefix addresses      
      self.__new_nodes_this_iter = 0
      peers_status = self.__get_active_nodes_summary_with_peers(current_network)
      
      # mark all nodes that are not online
      non_online = {
        x.get(self.const.PAYLOAD_DATA.NETMON_ADDRESS) : x.get(self.const.PAYLOAD_DATA.NETMON_EEID) for x in current_network.values() 
        if x.get(self.const.PAYLOAD_DATA.NETMON_STATUS_KEY, False) != self.const.DEVICE_STATUS_ONLINE
      }
      
      # mark all nodes that are not online
      for cached_addr in self.__allowed_nodes:
        if cached_addr in non_online and self.__allowed_nodes[cached_addr]["is_online"]:
          self.__allowed_nodes[cached_addr]["is_online"] = False
          self.P(f"[netmon_handler] Marking node '{non_online[cached_addr]}' <{cached_addr}> as offline.", color='r')
      # endfor marking non online nodes
      
      if self.__debug_netmon_count > 0:
        # self.P(f"NetMon debug:\n{self.json_dumps(self.__get_active_nodes(current_network), indent=2)}")
        self.P(f"[netmon_handler] peers status:\n{self.json_dumps(peers_status, indent=2)}")
        self.__check_dct_metadata()
        self.__debug_netmon_count -= 1
      #endif debug initial iterations
      
      for addr in peers_status:
        prefixed_addr = self.bc.maybe_add_prefix(addr)
        if prefixed_addr == self.ee_addr:
          # its us, no need to check whitelist
          continue
        if peers_status[addr]["allows_me"]:
          # we have found a whitelist that contains our address
          if addr not in self.__allowed_nodes:
            self.__allowed_nodes[addr] = {
              "whitelist" : peers_status[addr]["whitelist"],
              "last_config_get" : 0,
            } 
            self.__new_nodes_this_iter += 1
          #endif addr not in __allowed_nodes
          if not self.__allowed_nodes[addr].get("is_online", True):
            self.P("[netmon_handler] Node '{}' <{}> is back online.".format(peers_status[addr]["eeid"], addr))
          self.__allowed_nodes[addr]["is_online"] = True # by default we assume the node is online due to `__get_active_nodes_summary_with_peers`
        #endif addr allows me
      #endfor each addr in peers_status
      if self.__new_nodes_this_iter > 0:
        self.P(f"[netmon_handler] Found {self.__new_nodes_this_iter} new peered nodes.")
    #endif len(current_network) == 0
    return    
  

  def process(self):
    payload = None
    self.__maybe_send_requests()
    self.__maybe_send_configuration_to_allowed()
    self.__maybe_review_known()  
    return payload
  



import numpy as np

def _box_iou_batch(
	boxes_a: np.ndarray, boxes_b: np.ndarray
) -> np.ndarray:

  def box_area(box):
    return (box[2] - box[0]) * (box[3] - box[1])

  area_a = box_area(boxes_a.T)
  area_b = box_area(boxes_b.T)

  top_left = np.maximum(boxes_a[:, None, :2], boxes_b[:, :2])
  bottom_right = np.minimum(boxes_a[:, None, 2:], boxes_b[:, 2:])

  area_inter = np.prod(
  	np.clip(bottom_right - top_left, a_min=0, a_max=None), 2)
      
  return area_inter / (area_a[:, None] + area_b - area_inter)
  
def class_non_max_suppression(
   predictions: np.ndarray, iou_threshold: float = 0.5
) -> np.ndarray:
  rows, columns = predictions.shape

  sort_index = np.flip(predictions[:, 4].argsort())
  predictions = predictions[sort_index]

  boxes = predictions[:, :4]
  categories = predictions[:, 5]
  ious = _box_iou_batch(boxes, boxes)
  ious = ious - np.eye(rows)

  keep = np.ones(rows, dtype=bool)

  for index, (iou, category) in enumerate(zip(ious, categories)):
    if not keep[index]:
      continue

    condition = (iou > iou_threshold) & (categories == category)
    keep = keep & ~condition

  return keep[sort_index.argsort()]


def simple_nms(dets, thresh):
  """
  Non-maximum suppression. Returns indexes for kept boxes
  
  """
  y1 = dets[:, 0] # T
  x1 = dets[:, 1] # L
  y2 = dets[:, 2] # B
  x2 = dets[:, 3] # R
  scores = dets[:, 4]

  areas = (x2 - x1 + 1) * (y2 - y1 + 1)
  order = scores.argsort()[::-1]

  keep = []
  while order.size > 0:
    i = order[0]
    keep.append(i)
    xx1 = np.maximum(x1[i], x1[order[1:]])
    yy1 = np.maximum(y1[i], y1[order[1:]])
    xx2 = np.minimum(x2[i], x2[order[1:]])
    yy2 = np.minimum(y2[i], y2[order[1:]])

    w = np.maximum(0.0, xx2 - xx1 + 1)
    h = np.maximum(0.0, yy2 - yy1 + 1)
    intersection = w * h
    overlap = intersection / (areas[i] + areas[order[1:]] - intersection)

    inds = np.where(overlap <= thresh)[0]
    order = order[inds + 1]
  return keep
  

if __name__ == '__main__':
  # top1, left1, bottom1, right1 = [98, 1100, 265, 1252]
  # top2, left2, bottom2, right2 = [115, 1092, 377, 1253]
  
  top1, left1, bottom1, right1 = [116, 967, 238, 1103]
  top2, left2, bottom2, right2 = [126, 965, 383, 1104]

  
  top = max(top1, top2)
  left = max(left1, left2)
  right = min(right1, right2)
  bottom = min(bottom1, bottom2)
  
  h_overlap = bottom - top + 1
  w_overlap = right - left + 1
  
  area_overlap = h_overlap * w_overlap
  
  area_union = (right1 - left1 + 1) * (bottom1 - top1 + 1) + \
          (right2 - left2 + 1) * (bottom2 - top2 + 1) - \
          area_overlap
  iou = area_overlap / area_union
  print(area_overlap, area_union, area_overlap / area_union)
  
  data = np.array([
    [top1, left1, bottom1, right1, 69, 0],
    [top1+1, left1+2, bottom1+2, right1+2, 50, 1],
    [top2, left2, bottom2, right2, 22, 0],
  ])
  
  keep = simple_nms(data[:,:5], 0.5)
  print(keep)
  keep2 = class_non_max_suppression(data, 0.5)
  print(keep2)




from .continous_serving_process import ContinousServingProcess
from .base_serving_process import ModelServingProcess

from .basic_tf_session import BasicTFSessionServer
from .basic_tf_function import BasicTFFunctionServer
from .basic_th import UnifiedFirstStage
from .basic_th_anchor import BasicThAnchor



from naeural_core.business.default.web_app.naeural_fast_api_web_app import NaeuralFastApiWebApp

_CONFIG = {
  **NaeuralFastApiWebApp.CONFIG,

  # Only for the debug plugin
  "ASSETS": "extensions/business/xperimental",
  "DEBUG_MODE": False,

  'VALIDATION_RULES': {
    **NaeuralFastApiWebApp.CONFIG['VALIDATION_RULES'],
  },
}


class NaeuralFastApiDebugPlugin(NaeuralFastApiWebApp):
  """
  Debug plugin class for the Naeural Fast API Web App interface.
  """
  CONFIG = _CONFIG

  @NaeuralFastApiWebApp.endpoint(method='get', require_token=True)
  def get_stuff_with_token(self, token: str, stuff: str = ""):
    if token not in ['123', 'alabala']:
      return "Unauthorized token"
    stuff = stuff or {"empty": True}
    return {
      'Status': 'Success',
      'Stuff': stuff
    }

  @NaeuralFastApiWebApp.endpoint(method='get')
  def get_stuff_without_token(self, stuff: str = ""):
    stuff = stuff or {"empty": True}
    return {
      'Status': 'Success',
      'Stuff': stuff
    }







import torch as th
from .utils import get_activation
  
class GateLayer(th.nn.Module):
  def __init__(self, input_dim, output_dim, bias=-2):
    super().__init__()
    self.gate_lin = th.nn.Linear(input_dim, output_dim)
    self.gate_act = th.nn.Sigmoid()
    if bias is not None:
      self.gate_lin.bias.data.fill_(bias) # negative fc => zero sigmoid => (1 - 0) * bypass
    return
    
  def forward(self, inputs, value1, value2):
    th_gate = self.gate_lin(inputs)
    th_gate = self.gate_act(th_gate)
    th_out = th_gate * value1 + (1 - th_gate) * value2
    return th_out
  
  
class MultiGatedDense(th.nn.Module):
  """
  TODO: 
    - experiment with various biases for gates
  """
  def __init__(self, input_dim, output_dim, activ):
    super().__init__()
    self.bypass = th.nn.Linear(input_dim, output_dim, bias=False)

    self.fc = th.nn.Linear(input_dim, output_dim)
    self.fc_act = get_activation(activ)
    
    self.bn_pre = th.nn.BatchNorm1d(output_dim)
    self.bn_post = th.nn.BatchNorm1d(output_dim)

    self.bn_pre_vs_post_gate = GateLayer(input_dim, output_dim, bias=0) 

    self.lnorm_post = th.nn.LayerNorm(output_dim)

    self.bn_vs_lnorm = GateLayer(input_dim, output_dim, bias=0)
    
    self.has_bn_gate = GateLayer(input_dim, output_dim, bias=-1)
    
    self.final_gate = GateLayer(input_dim, output_dim, bias=-2)
    return

  
  def forward(self, inputs):
    # bypass
    th_bypass = self.bypass(inputs)
    
    # FC unit
    th_fc = self.fc(inputs)
    th_fc_act = self.fc_act(th_fc)
    
    # apply post layer norm
    th_fc_act_lnorm = self.lnorm_post(th_fc_act)    

    # FC with pre activ bn
    th_bn_pre = self.fc_act(self.bn_pre(th_fc))
    # FC with post activ bn
    th_bn_post = self.bn_post(th_fc_act)
    
    # select between bn pre or post
    th_bn_out = self.bn_pre_vs_post_gate(inputs, th_bn_pre, th_bn_post)
    
    # select between bn or layer norm
    th_bn_vs_lnorm = self.bn_vs_lnorm(inputs, th_bn_out, th_fc_act_lnorm)
    
    # select between normed or FC-activ
    th_norm_vs_simple = self.has_bn_gate(inputs, th_bn_vs_lnorm, th_fc_act)
    
    # finally select between processed or bypass
    th_out = self.final_gate(inputs, th_norm_vs_simple, th_bypass)
  
    return th_out
    
    
  

class GatedDense(th.nn.Module):
  def __init__(self, input_dim, output_dim, activ, bn):
    super().__init__()
    self.bn = bn
    self.linearity = th.nn.Linear(input_dim, output_dim, bias=not bn)
    if self.bn == 'pre':
      self.pre_bn = th.nn.BatchNorm1d(output_dim)
    elif self.bn == 'post':
      self.post_bn = th.nn.BatchNorm1d(output_dim)
    elif self.bn == 'lnorm':
      self.post_lnorm = th.nn.LayerNorm(output_dim)
    self.activ = get_activation(activ)
    self.gate = GateLayer(input_dim, output_dim)
    self.bypass_layer = th.nn.Linear(input_dim, output_dim, bias=False)
  
  def forward(self, inputs):
    # the skip/residual
    th_bypass = self.bypass_layer(inputs)
    # the normal linear-bn-activ
    th_x = self.linearity(inputs)
    if self.bn == 'pre':
      th_x = self.pre_bn(th_x)
    th_x = self.activ(th_x)
    if self.bn == 'post':
      th_x = self.post_bn(th_x)
    elif self.bn == 'lnorm':
      th_x = self.post_lnorm(th_x)
    # the gate
    th_x = self.gate(inputs, th_x, th_bypass)
    return th_x
  



from naeural_core.serving.base.basic_th import UnifiedFirstStage as ParentServingProcess
from naeural_core.local_libraries.nn.th.utils import th_resize_with_pad

_CONFIG = {
  **ParentServingProcess.CONFIG,

  "MODEL_CLASSES_FILENAME": "coco.txt",
  "URL_CLASS_NAMES": None,  # "minio:Y5/coco.txt",

  "MODEL_WEIGHTS_FILENAME": "20230723_y8s_nms.ths",
  "MODEL_WEIGHTS_FILENAME_DEBUG": "20230723_y8s_nms_top6.ths",

  "URL": "minio:Y8/20230723_y8s_nms.ths",
  "URL_DEBUG": "minio:Y8/20230723_y8s_nms_top6.ths",

  'IMAGE_HW': (448, 640),
  "DEBUG_SERVING": False,

  "PICKED_INPUT": "IMG",
  
  "COLOR_TAGGING": False,

  'MAX_BATCH_FIRST_STAGE': 8,

  'VALIDATION_RULES': {
    **ParentServingProcess.CONFIG['VALIDATION_RULES'],
  },
}

DEBUG_COVERED_SERVERS = False

__VER__ = '0.2.0.0'


class YfBase(ParentServingProcess):
  CONFIG = _CONFIG

  def __init__(self, **kwargs):
    ver = kwargs.pop('version', None)
    if ver is None:
      ver = __VER__

    # this must be correctly configured by each individual serving process
    self._has_second_stage_classifier = False
    self.resized_input_images = None
    self.original_input_images = None
    super(YfBase, self).__init__( **kwargs)
    return

  @property
  def get_model_weights_filename(self):
    if self.cfg_debug_serving:
      return self.cfg_model_weights_filename_debug
    return self.cfg_model_weights_filename

  @property
  def get_url(self):
    if self.cfg_debug_serving:
      return self.cfg_url_debug
    return self.cfg_url

  @property
  def has_second_stage_classifier(self):
    return self._has_second_stage_classifier and self.server_name == self.predict_server_name
  
  

  def _get_model(self, config):
    if not isinstance(config, dict):
      config = {
        'ths' : (config, None)
      }
    #endif check for torchscript file config
    model, model_loaded_config, fn = self.prepare_model(
      backend_model_map=config,
      forced_backend=self.cfg_backend,
      post_process_classes=True,
      return_config=True,
      batch_size = self.cfg_max_batch_first_stage
    )
    # FIXME: is the model file path appropriate here to use as the config key?
    self.graph_config[fn] = model_loaded_config
    return model

  def _pre_process_images(self, images, return_original=False, normalize_original=False, half_original=False, **kwargs):
    lst_original_shapes = []
    # run GPU based pre-processing
    self._start_timer('resize_in_gpu')
    results = th_resize_with_pad(
      img=images,
      h=self.cfg_input_size[0],
      w=self.cfg_input_size[1],
      device=self.dev,
      normalize=True,  # Why below ?
      return_original=return_original,
      normalize_original=normalize_original,
      half_original=half_original,
      half=self.cfg_fp16
    )
    if len(results) < 3:
      prep_inputs, lst_original_shapes = results
    else:
      prep_inputs, lst_original_shapes, lst_original_images = results
      self.original_input_images = lst_original_images
    # endif check for results length
    
    self.resized_input_images = prep_inputs
    self._stop_timer('resize_in_gpu')

    self._lst_original_shapes = lst_original_shapes
    return prep_inputs

  def _make_crop(self, frame, ltrb, original_shape, return_offsets=False):
    l, t, r, b = self.scale_coords(
      img1_shape=self.cfg_input_size,
      coords=self.deepcopy(ltrb),
      img0_shape=original_shape,
    ).int().flatten().clamp(min=0)

    cropped_image = self.tv.transforms.functional.crop(
      img=frame,
      left=l.clamp(min=0),
      top=t.clamp(min=0),
      width=(r - l).clamp(min=4) + 1,
      height=(b - t).clamp(min=4) + 1,
    )

    return cropped_image if not return_offsets else (cropped_image, (l, t))

  def _second_stage_process(self, th_preds, th_inputs=None, **kwargs):
    th_pred_nms, th_n_dets = th_preds
    self.log.start_timer('slice_det')
    pred_nms = [x[:th_n_dets[i]] for i, x in enumerate(th_pred_nms)]
    self.log.stop_timer('slice_det')

    pred_second = None
    if self.has_second_stage_classifier:
      self.log.start_timer('2nd_stage_clf')
      pred_nms_second_stage = [x[:, :6] for x in pred_nms]
      pred_second = self._second_stage_classifier(pred_nms_second_stage, th_inputs)
      self.log.stop_timer('2nd_stage_clf')
    # endif

    self.log.start_timer('yolo_to_cpu')
    pred_nms_cpu = [x.cpu().numpy() for x in pred_nms]
    self.log.stop_timer('yolo_to_cpu')

    return pred_nms_cpu, pred_second

  def _aggregate_batch_predict(self, lst_results):
    th_preds_batch = self.th.cat([x[0] for x in lst_results])
    th_preds_n_det = self.th.cat([x[1] for x in lst_results])

    return th_preds_batch, th_preds_n_det

  def _post_process(self, preds):
    pred_nms_cpu, _ = preds
    nr_images = len(self._lst_original_shapes)
    lst_results = []
    for i in range(nr_images):
      # now we have each individual image and we generate all objects
      # what we need to do is to match `second_preds` to image id & then
      # match second clf with each box
      np_pred_nms_cpu = pred_nms_cpu[i]
      original_shape = self._lst_original_shapes[i]
      np_pred_nms_cpu[:, :4] = self.scale_coords(
        img1_shape=self.cfg_input_size,
        coords=np_pred_nms_cpu[:, :4],
        img0_shape=original_shape,
      ).round()
      lst_inf = []
      for det in np_pred_nms_cpu:
        det = [float(x) for x in det]
        # order is [left, top, right, bottom, proba, class] => [L, T, R, B, P, C, RP1, RC1, RP2, RC2, RP3, RC3]
        L, T, R, B, P, C = det[:6]
        # now check if color needs to be computed
        # the color tagging is done either in the serving or in the business logic
        # depending on the configuration and the target compute overhead
        # for heavy inference it is better to do it in the business logic
        color_tag = None
        if self.cfg_color_tagging:
          pass
        # endif compute color
        dct_obj = {
          self.consts.TLBR_POS: [int(T), int(L), int(B), int(R)],
          self.consts.PROB_PRC: round(float(P), 2),
          self.consts.TYPE: self.class_names[int(C)] if self.class_names is not None else C,          
          # additional properties such as color, etc.
          self.consts.COLOR_TAG: color_tag,
        }
        if self.cfg_debug_serving:
          dct_obj['CANDIDATES'] = [
            [x, self.class_names[int(y)] if self.class_names is not None else y]
            for (x, y) in list(zip(det[6:][::2], det[6:][1::2]))
          ]
        lst_inf.append(dct_obj)
      lst_results.append(lst_inf)
    return lst_results



from naeural_core import constants as ct
import json

class _ConfigManagerCheckMixin(object):
  def __init__(self):
    super(_ConfigManagerCheckMixin, self).__init__()
    return

  def __append_exception_message(self, messages, str_message, extra=None):
    if extra is not None:
      str_message += '\n{}'.format(json.dumps(extra))

    messages.append({'notif' : ct.STATUS_TYPE.STATUS_EXCEPTION, 'msg' : str_message})
    self.P(str_message, color='r')
    return

  def _check_config_app(self, config_app):
    """
    Receives an application config, checks it and maybe cleans it in order to start the application.

    Parameters:
    ----------
    config_app: dict, mandatory
      application config

    Returns:
    -------
    config_app: dict
      the maybe cleaned application config
    """
    messages = []
    can_start_app = True

    # check first if 'COMMUNICATION' is configured
    if 'COMMUNICATION' not in config_app:
      crt_msg = "ERROR! 'COMMUNICATION' not configured in config app."
      self.__append_exception_message(messages, crt_msg)
      can_start_app = False
    else:
      _params = config_app['COMMUNICATION'].get('PARAMS', {})
      _type = config_app['COMMUNICATION'].get('TYPE', '')
      _instances = config_app['COMMUNICATION'].get('INSTANCES', {})

      if not bool(_params):
        crt_msg = "ERROR! 'PARAMS' dictionary not configured in config_app's 'COMMUNICATION' zone"
        self.__append_exception_message(messages, crt_msg)
        can_start_app = False

      if not bool(_type):
        crt_msg = "ERROR! 'TYPE' not configured in config_app's 'COMMUNICATION' zone"
        self.__append_exception_message(messages, crt_msg)
        can_start_app = False

      if not bool(_instances):
        crt_msg = "ERROR! 'INSTANCES' not configured in config_app's 'COMMUNICATION' zone"
        self.__append_exception_message(messages, crt_msg)
        can_start_app = False
    #endif


    for msg in messages:
      self._create_notification(**msg)

    if not can_start_app:
      raise RuntimeError("Application cannot start. Please see above errors.")

    return config_app

  def _check_config_stream(self, config_stream):
    """
    Receives a stream configuration, does some checks and determines whether the stream configuration is good or not

    Parameters:
    ----------
    config_stream: dict, mandatory
      the stream configuration

    Returns:
    --------
    is_good_stream: bool
      Whether the stream configuration is good or not
    """
    messages = []
    is_good_stream = True
    stream_name = config_stream['NAME']

    stream_type = config_stream.get('TYPE', None)
    if stream_type is None:
      crt_msg = "WARNING! 'TYPE' not configured for stream `{}`. Dropping stream `{}`".format(stream_name, stream_name)
      self.__append_exception_message(messages, crt_msg)
      is_good_stream = False
    #endif

    for msg in messages:
      self._create_notification(
        initiator_id=config_stream.get(ct.PAYLOAD_DATA.INITIATOR_ID),
        session_id=config_stream.get(ct.PAYLOAD_DATA.SESSION_ID),
        stream_name=stream_name,
        **msg
      )

    return is_good_stream

  def _check_config_plugin(self, stream_name, config_plugin):
    """
    Receives a plugin configuration, does some checks and determines whether the plugin configuration is good or not

    Parameters:
    ----------
    stream_name: str, mandatory
      the name of the stream where the plugin is configured

    config_plugin: dict, mandatory
      the plugin configuration

    Returns:
    --------
    is_good_plugin: bool
      Whether the plugin configuration is good or not
    """
    messages = []
    is_good_plugin = True

    if 'SIGNATURE' not in config_plugin:
      crt_msg = "WARNING! 'SIGNATURE' not configured for plugin in stream `{}`. Dropping plugin".format(stream_name)
      self.__append_exception_message(messages, crt_msg, config_plugin)
      is_good_plugin = False

    lst_config_instances = config_plugin.get('INSTANCES', [])
    if len(lst_config_instances) == 0:
      crt_msg = "WARNING! 'INSTANCES' not configured for plugin in stream `{}`. Dropping plugin".format(stream_name)
      self.__append_exception_message(messages, crt_msg, config_plugin)
      is_good_plugin = False

    for msg in messages:
      self._create_notification(
        stream_name=stream_name,
        **msg
      )

    return is_good_plugin

  def _check_config_instance(self, stream_name, plugin_signature, config_instance):
    """
    Receives a plugin instance configuration, does some checks and determines whether the plugin instance configuration is good or not

    Parameters:
    ----------
    stream_name: str, mandatory
      the name of the stream where the plugin instance is configured

    plugin_signature: str, mandatory
      the signature of the plugin

    config_instance: dict, mandatory
      the plugin instance configuration

    Returns:
    --------
    is_good_instance: bool
      Whether the plugin instance configuration is good or not
    """
    messages = []
    is_good_instance = True

    if False:
      self.P("Checking config instance for Stream:'{}' -> Plugin:'{}' -> Instance config:{}".format(
        stream_name, plugin_signature, 
        '\n{}'.format(self.log.dict_pretty_format(config_instance)) if isinstance(config_instance, dict) else "'{}'".format(config_instance),
        ),
        color='m'
      )


    
    if 'INSTANCE_ID' not in config_instance:
      config_instance['INSTANCE_ID'] = self.log.now_str()      
    
    if 'INSTANCE_ID' not in config_instance:
      crt_msg = "WARNING! 'INSTANCE_ID' not configured in plugin `{}`, stream `{}`. Dropping instance".format(plugin_signature, stream_name)
      self.__append_exception_message(messages, crt_msg, config_instance)
      is_good_instance = False

    points = config_instance.get('POINTS', [])

    if len(points) == 0:
      coords = "NONE"
    elif isinstance(points[0], int):
      coords = "TLBR"
    elif isinstance(points[0], list):
      coords = "POINTS"
    else:
      crt_msg = "WARNING! Cannot deduce coords type given the points {} in plugin `{}`, stream `{}`. Dropping instance".format(points, plugin_signature, stream_name)
      self.__append_exception_message(messages, crt_msg, config_instance)
      is_good_instance = False
      coords = None
    #endif

    if coords == 'TLBR':
      if len(points) != 4:
        crt_msg = "WARNING! 'POINTS' bad configured for COORDS=TLBR in plugin `{}`, stream `{}`. Dropping instance".format(plugin_signature, stream_name)
        self.__append_exception_message(messages, crt_msg, config_instance)
        is_good_instance = False
      elif None in points:
        crt_msg = "WARNING! 'POINTS' bad configured for COORDS=TLBR (it contains None) in plugin `{}`, stream `{}`. Dropping instance".format(plugin_signature, stream_name)
        self.__append_exception_message(messages, crt_msg, config_instance)
        is_good_instance = False
      #endif
    #endif

    if coords == 'POINTS':
      if not all([len(x) == 2 for x in points]):
        crt_msg = "WARNING! 'POINTS' bad configured for COORDS=POINTS in plugin `{}`, stream `{}`. Dropping instance".format(plugin_signature, stream_name)
        self.__append_exception_message(messages, crt_msg, config_instance)
        is_good_instance = False
      elif None in self.log.flatten_2d_list(points):
        crt_msg = "WARNING! 'POINTS' bad configured for COORDS=POINTS (it contains None) in plugin `{}`, stream `{}`. Dropping instance".format(plugin_signature, stream_name)
        self.__append_exception_message(messages, crt_msg, config_instance)
        is_good_instance = False
      #endif
    #endif

    for msg in messages:
      self._create_notification(
        stream_name=stream_name,
        **msg
      )

    return is_good_instance

  def _keep_plugin_good_instances(self, stream_name, config_plugin):
    """
    Keeps only the good instances inside a plugin configuration

    Parameters:
    ----------
    stream_name: str, mandatory
      the name of the stream where the plugin is configured

    config_plugin: dict, mandatory
      the plugin configuration

    Returns:
    -------
    config_plugin: dict
      the cleaned plugin configuration
    """
    if not self._check_config_plugin(stream_name, config_plugin):
      return

    messages = []
    signature = config_plugin['SIGNATURE']
    lst_config_instances = config_plugin['INSTANCES']
    new_lst_config_instances = []

    lst_instance_ids = []
    for config_instance in lst_config_instances:
      is_good_instance = self._check_config_instance(stream_name, signature, config_instance)
      if is_good_instance:
        instance_id = config_instance['INSTANCE_ID']
        if instance_id in lst_instance_ids:
          crt_msg = "WARNING! Instance id {} already existing for plugin `{}` stream `{}`. Keeping just the first one".format(instance_id, signature, stream_name)
          self.__append_exception_message(messages, crt_msg)
        else:
          lst_instance_ids.append(instance_id)
          new_lst_config_instances.append(config_instance)
        #enif
      #endif
    #endfor

    config_plugin['INSTANCES'] = new_lst_config_instances

    for msg in messages:
      self._create_notification(**msg)

    ### maybe it becomes bad after filtering instances
    if not self._check_config_plugin(stream_name, config_plugin):
      return

    return config_plugin
  
  
  def maybe_add_mandatory_plugin_to_stream(self, dct_config_stream):
    """
    Add default mandatory plugin to stream if the plugin does not already exists 
    and if stream requires it.

    Parameters
    ----------
    dct_config_stream : dict
      the pipeline config.

    Returns
    -------
    None.

    """
    # direct connect to config startup
    dct_capture_env = self.owner.cfg_capture_environment
    # get default
    default = dct_capture_env.get(ct.CONFIG_STREAM.DEFAULT_PLUGIN, False)
    # det and overwrite from given stream
    def_sig = dct_config_stream.get(ct.CONFIG_STREAM.DEFAULT_PLUGIN, default)
    # below check code is written in "explicit" mode
    if def_sig is None:
      # no valid default signature
      return
    elif isinstance(def_sig, bool):
      if def_sig:
        def_sig = ct.CONFIG_STREAM.DEFAULT_PLUGIN_SIGNATURE
      else:
        return
    elif isinstance(def_sig, str):
      if len(def_sig) <= 1:  
        # obviously invalid signature
        return
    else:
      # unknown value
      return
    default_plugin_config = dct_config_stream.get(ct.CONFIG_STREAM.DEFAULT_PLUGIN_CONFIG, {})
    stream_name = dct_config_stream[ct.NAME]
    stream_type = dct_config_stream[ct.TYPE].upper()
    additional_config = {}
    if stream_type in ct.CONFIG_STREAM.NO_DATA_STREAMS:
      additional_config = {
        ct.BIZ_PLUGIN_DATA.ALLOW_EMPTY_INPUTS : True,
        ct.BIZ_PLUGIN_DATA.RUN_WITHOUT_IMAGE  : True,
        ct.BIZ_PLUGIN_DATA.PROCESS_DELAY      : 0.5,
      }
    lst_config_plugins = dct_config_stream.get(ct.PLUGINS, [])
    lst_signatures = [x[ct.SIGNATURE].upper() for x in lst_config_plugins]
    has_signature = def_sig.upper() in lst_signatures
    if not has_signature:
      self.P("Adding mandatory plugin '{}' to stream '{}'".format(def_sig, stream_name))
      if not isinstance(dct_config_stream.get(ct.PLUGINS), list):
        dct_config_stream[ct.PLUGINS] = []
      short = self.log.get_short_name(def_sig)
      dct_config_stream[ct.PLUGINS].append({
        ct.SIGNATURE : def_sig, # signature of default plugin - a custom exec type
        ct.INSTANCES : [{
          ct.BIZ_PLUGIN_DATA.INSTANCE_ID                 : short + '_default',
          # ct.BIZ_PLUGIN_DATA.MAX_INPUTS_QUEUE_SIZE       : 1, # should be 1 or more ?
          **additional_config,
          **default_plugin_config
        }]
      })
    return
      
  
  def _get_plugin_instance(self, stream_name, signature, instance_id):
    """
    Returns the plugin instance configuration given the stream name, plugin signature and instance id
    """
    config_stream = self.dct_config_streams.get(stream_name)
    if config_stream is not None:
      lst_config_plugins = config_stream.get('PLUGINS', [])
      for config_plugin in lst_config_plugins:
        if config_plugin['SIGNATURE'].upper() == signature.upper():
          lst_instances = config_plugin['INSTANCES']
          for instance in lst_instances:
            if instance['INSTANCE_ID'] == instance_id:
              return instance
    return
          
      

  def _keep_stream_good_plugins(self, config_stream):
    """
    Keeps only the good plugins inside a stream configuration

    Parameters:
    ----------
    config_stream: dict, mandatory
      the stream configuration

    Returns:
    -------
    config_stream: dict
      the cleaned stream configuration
    """
    if not self._check_config_stream(config_stream):
      return

    messages = []

    new_lst_config_plugins = []
    stream_name = config_stream[ct.CONFIG_STREAM.NAME]
    pipeline_type = config_stream[ct.CONFIG_STREAM.TYPE]
    allowed_plugins = config_stream.get(ct.CONFIG_STREAM.ALLOWED_PLUGINS, [])
    has_allowed_plugins = len(allowed_plugins) > 0
    lst_config_plugins = config_stream.get(ct.CONFIG_STREAM.PLUGINS, [])

    lst_plugin_signatures = []
    for config_plugin in lst_config_plugins:
      new_config_plugin = self._keep_plugin_good_instances(stream_name, config_plugin)
      if new_config_plugin is not None:
        plugin_signature = new_config_plugin[ct.PAYLOAD_DATA.SIGNATURE]
        if has_allowed_plugins and plugin_signature not in allowed_plugins:
          crt_msg = "WARNING! plugin `{}` not allowed for pipeline `{}` ({}). Dropping plugin".format(
            plugin_signature, stream_name, pipeline_type
          )
          self.P(crt_msg, color='error')
          self.__append_exception_message(messages, crt_msg)
          continue
        #endif allowed plugins section
        
        if plugin_signature in lst_plugin_signatures:
          for existing in lst_config_plugins:
            if existing[ct.PLUGIN_INFO.SIGNATURE] == plugin_signature:
              break
          instances = [x[ct.PLUGIN_INFO.INSTANCE_ID] for x in existing[ct.CONFIG_STREAM.INSTANCES]]
          new_instances = [x[ct.PLUGIN_INFO.INSTANCE_ID] for x in new_config_plugin[ct.CONFIG_STREAM.INSTANCES]]
          crt_msg = "WARNING! plugin `{}`:{} already exists for stream `{}` with instances: {}. Keeping just the first one:\n{}".format(
            plugin_signature, new_instances, stream_name, instances,
            lst_config_plugins,
          )
          self.__append_exception_message(messages, crt_msg)
        else:
          lst_plugin_signatures.append(plugin_signature)
          new_lst_config_plugins.append(new_config_plugin)
        #endif existing plugin or new plugin
      #endif plugin valid
    #endfor

    config_stream[ct.CONFIG_STREAM.PLUGINS] = new_lst_config_plugins

    for msg in messages:
      self._create_notification(
        initiator_id=config_stream.get(ct.PAYLOAD_DATA.INITIATOR_ID),
        session_id=config_stream.get(ct.PAYLOAD_DATA.SESSION_ID),
        stream_name=stream_name,
        **msg
      )

    if not self._check_config_stream(config_stream):
      return
    
    self.log.P("Config manager preparing to start stream '{}' with following plugins: {}".format(
      stream_name, lst_plugin_signatures), color=ct.COLORS.DCT,
    )

    return config_stream

  def keep_good_stream(self, config_stream):
    """
    Determines whether a stream configuration is good or not; if it is good, then the method cleans it and
    finally adds default information/data to the stream.
    Public method to be called when an individual stream comes in the box.
    

    Parameters:
    -----------
    config_stream: dict, mandatory
      the stream configuration

    Returns:
    -------
    None or config_stream: dict
    """
    is_good_stream = self._check_config_stream(config_stream)
    if not is_good_stream:
      return

    config_stream = self._keep_stream_good_plugins(config_stream)
    
    
    self.maybe_add_mandatory_plugin_to_stream(dct_config_stream=config_stream)
    return config_stream

  def keep_all_good_streams(self, dct_config_streams):
    """
    Takes all streams configurations and calls `self.keep_good_stream` on them.
    Public method to be called when an multiple stream comes in the box.

    Parameters:
    ----------
    dct_config_streams: dict, mandatory
      key: value where each key is the name of a stream and value is the stream configuration

    Returns:
    --------
    new_dct_config_streams: dict
      The new streams after cleaning and filtering
    """
    self.log.start_timer('keep_all_good_streams')
    new_dct_config_streams = {}
    for stream_name, config_stream in dct_config_streams.items():
      new_config_stream = self.keep_good_stream(config_stream)
      if new_config_stream is not None:
        new_dct_config_streams[stream_name] = new_config_stream
      #endif
    #endfor

    self.log.stop_timer('keep_all_good_streams', skip_first_timing=False)
    return new_dct_config_streams



import torch as th
from naeural_core.local_libraries.nn.th.layers import (
  ConcatenateLayer,
  Conv2dExt,
  DSepConv2DModule,
  DSepConv2DSkipModule,
  ConvResizer,
  IdentityLayer,
  EmbeddingTransform,
)
from naeural_core.local_libraries.nn.utils import get_dropout_rate
from naeural_core.local_libraries.nn.th.utils import l2distance, get_optimizer_class, get_activation, get_dropout


class CNNColumn(th.nn.Module):
  """
  Convolutional Neural Network Column class that applies a series of convolutional layers.
  """
  def __init__(
      self, lst_filters, layer_norm,
      in_channels=None, input_dim=None,
      nconv=1, stride_multiplier=1,
      act='relu', dropout=0, dropout_type='classic',
      dropout_value_type='constant',
      skipping='residual', resize_conv=False
  ):
    """
    Parameters
    ----------
    lst_filters : list(tuple(int, int)), list of tuples defining the number of filters and kernel size for each layer
    layer_norm : bool, whether to apply layer normalization
    in_channels : int or list or tuple, number of input channels
    input_dim : tuple of int, input dimensions
    nconv : int, number of times to apply a certain convolutional layer
    stride_multiplier : int, stride multiplier for the first layer
    act : str, activation function
    dropout : float, dropout rate
    dropout_type : str, type of dropout
    dropout_value_type : str, type of dropout value
    skipping : str, type of skipping connection
    resize_conv : bool, whether to apply a convolutional resizer
    """
    super().__init__()
    if (input_dim is None or not all(input_dim)) and layer_norm is True:
      raise ValueError("Cannot apply layer norm without knowing input shapes")

    if input_dim is None or not any(input_dim):
      input_dim = (in_channels, None, None)

    assert any(input_dim), ValueError("CNNColumn expects 'in_channels' or 'input_dim' parameters")

    if not isinstance(act, th.nn.Module):
      act = get_activation(act)
    # endif
    self.input_dim = input_dim
    self.residual = False if skipping == 'skip' else True
    self.backbone = self.backbone_skip_forward if skipping == 'both' else self.backbone_forward
    conv_module = DSepConv2DSkipModule if skipping == 'both' else DSepConv2DModule
    reduced_dim = None

    self.column = th.nn.ModuleList()
    for i in range(len(lst_filters)):
      dropout_rate = get_dropout_rate(
        dropout=dropout,
        step=i,
        max_step=len(lst_filters) - 1,
        value_type=dropout_value_type
      )
      self.column.append(
        conv_module(
          input_dim=input_dim,
          out_channels=lst_filters[i][0],
          n_convs=nconv,
          kernel_size=lst_filters[i][1],
          stride=lst_filters[i][2] if len(lst_filters[i]) > 2 else (
            lst_filters[i][1] * stride_multiplier if i == 0 else 1),
          dropout=dropout_rate,
          dropout_type=dropout_type,
          activation=act,
          patching=False
        )
      )
      input_dim = self.column[-1].output_dim
      if reduced_dim is None:
        reduced_dim = input_dim
      # endif reduced_dim
    # endfor
    if skipping == 'both':
      self.linear_resizer = ConvResizer(
        input_dim=reduced_dim,
        output_dim=input_dim,
        apply_conv=resize_conv
      )
      input_dim = (input_dim[0] + self.linear_resizer.out_channels, *input_dim[1:])
    else:
      self.linear_resizer = None
    # endif skipping

    self.output_dim = input_dim
    if layer_norm:
      self.layer_norm = th.nn.LayerNorm(input_dim)
    else:
      self.layer_norm = IdentityLayer()
    # endif layer_norm
    return

  def backbone_forward(self, th_x):
    for layer in self.column:
      th_x = layer(th_x)

    return th_x

  def backbone_skip_forward(self, th_x):
    th_x, th_x_reduced = self.column[0](th_x)
    for layer in self.column[1:]:
      th_x, _ = layer(th_x)
    th_x_reduced = self.linear_resizer(th_x_reduced)

    return th.cat([th_x, th_x_reduced], dim=1)

  def forward(self, th_x):
    th_x = self.backbone(th_x)

    th_x = self.layer_norm(th_x)

    return th_x

  def __repr__(self):
    s = super().__repr__()

    if self.input_dim is None:
      return s

    s += " [Input={} => Output={}]".format(
      self.input_dim,
      self.output_dim
    )
    return s


class ImageEncoder(th.nn.Module):
  def __init__(
      self, filters, stride_multiplier, cnn_act, layer_norm, nconv,
      in_channels=None, embedding_transform='gmp/gap', input_dim=None,
      dropout=0, dropout_type='classic', dropout_value_type='constant',
      resize_conv=False, skipping='residual', equalize_columns=False,
      equalize_method='min'
  ):
    """
    TODO: Docstrings

    TODO: Implement output_dim
    """
    super().__init__()

    if input_dim is None or not any(input_dim):
      input_dim = (in_channels, None, None)

    assert any(input_dim), ValueError("ImageEncoder expects 'in_channels' or 'input_dim' parameters")
    self.input_dim = input_dim

    self.output_filters = 0
    columns = []
    for i in range(len(filters)):
      tf_x_column = CNNColumn(
        input_dim=input_dim,
        lst_filters=filters[i],
        act=cnn_act,
        stride_multiplier=stride_multiplier,
        layer_norm=layer_norm,
        nconv=nconv,
        dropout=dropout,
        dropout_type=dropout_type,
        dropout_value_type=dropout_value_type,
        skipping=skipping,
        resize_conv=resize_conv
      )
      transform = EmbeddingTransform(
        type=embedding_transform,
        input_dim=tf_x_column.output_dim
      )
      columns.append([tf_x_column, transform])
    # endfor columns
    embedding_sizes = [transform.output_dim for _, transform in columns]
    if equalize_columns:
      # Equalize the number of output filters for each column
      # First, decide the embedding size
      if equalize_method == 'min':
        embedding_size = min(embedding_sizes)
      elif equalize_method == 'max':
        embedding_size = max(embedding_sizes)
      else:
        raise ValueError(f"Equalize method {equalize_method} not implemented")
      # endif equalize_method
      # Second, add a linear layer to each column to equalize the number of filters
      for i in range(len(columns)):
        columns[i].append(th.nn.Linear(embedding_sizes[i], embedding_size))
      # endfor columns
      self.output_filters = len(columns) * embedding_size
    else:
      self.output_filters = sum(embedding_sizes)
    # endif equalize_columns
    columns = [th.nn.Sequential(*column) for column in columns]
    self.columns = th.nn.ModuleList(columns)
    # TODO: fix verbosity
    return

  def forward(self, x):
    results = []

    # x  self.stem(x)

    for column in self.columns:
      results.append(column(x))

    results = th.cat(results, dim=1)
    return results

  @property
  def output_dim(self):
    return self.output_filters

  def __repr__(self):
    s = super().__repr__()

    if self.input_dim is None:
      return s

    s += " [Input={} => Output={}]".format(
      self.input_dim,
      self.output_dim
    )
    return s


class StemLayer(th.nn.Module):
  def __init__(
    self, filters, in_channels=None, input_dim=None, act='relu', skip_last_act=False, **kwargs
  ):
    """
    Neural Network Stem Layer class that applies a series of convolutional layers
    for the initial part of the network with the purpose of feature extraction.
    Parameters
    ----------
    filters : list(int) or list(tuple(int)), list of filters for each layer
    in_channels : int, number of input channels
    input_dim : tuple of int, input dimensions
    act : str, activation function
    skip_last_act : bool, whether to skip the last activation function
    kwargs : dict, additional parameters
    """
    super().__init__()
    if input_dim is None or not any(input_dim):
      input_dim = (in_channels, None, None)
    assert any(input_dim), ValueError("StemLayer expects 'in_channels' or 'input_dim' parameters")

    self.input_dim = input_dim
    current_input_shape = input_dim
    stem_layers = []
    for i, layer_filter in enumerate(filters):
      if isinstance(layer_filter, int):
        layer_filter = [layer_filter]
      current_layer = Conv2dExt(
        input_shape=current_input_shape,
        in_channels=current_input_shape[0],
        out_channels=layer_filter[0],
        kernel_size=layer_filter[1] if len(layer_filter) > 1 else 3,
        stride=layer_filter[2] if len(layer_filter) > 2 else 2
      )
      stem_layers.append(current_layer)
      if not skip_last_act or i < len(filters) - 1:
        stem_layers.append(get_activation(act))
      current_input_shape = current_layer.output_shape
    # endfor layers
    self.stem = th.nn.Sequential(*stem_layers)
    self.output_dim = current_input_shape
    return

  def forward(self, x):
    return self.stem(x)

  def __repr__(self):
    s = super().__repr__()

    if self.input_dim is None:
      return s

    s += " [Input={} => Output={}]".format(
      self.input_dim,
      self.output_dim
    )
    return s


class ImageEncoderWithStem(th.nn.Module):
  def __init__(self, filters, stride_multiplier, cnn_act, layer_norm, nconv, stem_layer, use_stem_activation, in_channels=None, embedding_transform=None, input_dim=None):
    """
    TODO: Docstrings

    TODO: Implement output_dim
    """
    super().__init__()

    if input_dim is None or not any(input_dim):
      input_dim = (in_channels, None, None)

    assert any(input_dim), ValueError("ImageEncoderWithStem expects 'in_channels' or 'input_dim' parameters")

    self.input_dim = input_dim
    stem_layers = []
    stem_layers.append(
      Conv2dExt(
        input_shape=input_dim,

        in_channels=input_dim[0],
        out_channels=stem_layer[0],
        kernel_size=stem_layer[1],
        stride=stem_layer[2] if len(stem_layer) > 2 else 2,
      )
    )

    if use_stem_activation:
      stem_layers.append(th.nn.ReLU())
    self.stem = th.nn.Sequential(*stem_layers)

    self.output_filters = 0
    columns = []
    for i in range(len(filters)):
      tf_x_column = CNNColumn(
        input_dim=stem_layers[0].output_shape,
        lst_filters=filters[i],
        act=cnn_act,
        stride_multiplier=stride_multiplier,
        layer_norm=layer_norm,
        nconv=nconv
      )
      transform = EmbeddingTransform(
          type=embedding_transform,
          input_dim=tf_x_column.output_dim
        )
      columns.append(th.nn.Sequential(
        tf_x_column,
        transform
      ))
      self.output_filters += transform.output_dim
    self.columns = th.nn.ModuleList(columns)

  def forward(self, x):
    results = []

    x = self.stem(x)

    for column in self.columns:
      results.append(column(x))

    results = th.cat(results, dim=1)
    return results

  @property
  def output_dim(self):
    return self.output_filters

  def __repr__(self):
    s = super().__repr__()

    if self.input_dim is None:
      return s

    s += " [Input={} => Output={}]".format(
      self.input_dim,
      self.output_dim
    )
    return s


class ReadoutConv(th.nn.Module):
  """
  Readout Convolutional Neural Network class that applies a series of depth separable convolutional layers.
  """
  def __init__(
      self, lst_convs, in_channels, input_dim,
      act='relu6', dropout=0, embedding_type='flatten',
      dropout_type='classic', dropout_value_type='constant'
  ):
    """

    Parameters
    ----------
    lst_convs : list(int), list of number of filters for each layer
    in_channels : int, number of input channels
    input_dim : tuple of int, input dimensions
    act : str, activation function
    dropout : float, dropout rate
    embedding_type : str, type of embedding
    dropout_type : str, type of dropout
    dropout_value_type : str, type of dropout value applied after each depthwise separable convolutional layer.
      This is set to '' by default for backward compatibility.
    """
    super(ReadoutConv, self).__init__()

    self.lst_convs = lst_convs

    lst_blocks = []
    self.output_dim = input_dim
    self.in_channels = in_channels
    for step, out_channels in enumerate(lst_convs):
      dropout_rate = get_dropout_rate(
        dropout=dropout,
        step=step,
        max_step=len(lst_convs) - 1,
        value_type=dropout_value_type
      )
      conv = DSepConv2DModule(
        in_channels=in_channels,
        out_channels=out_channels,
        n_convs=1,
        kernel_size=3,
        input_dim=input_dim,
        activation=get_activation(act),
        patching=False,
        stride=1,
        dropout=dropout_rate,
        dropout_type=dropout_type
      )
      input_dim = conv.output_dim
      self._output_dim = conv.output_dim
      lst_blocks.append(conv)
      in_channels = out_channels
    # endfor

    self.blocks = th.nn.Sequential(*lst_blocks)
    self.embedding = EmbeddingTransform(
      type=embedding_type,
      input_dim=self._output_dim
    )
    self.dropout_rate = dropout
    self.dropout = get_dropout(dropout=self.dropout_rate, dropout_type=dropout_type)
    return

  @property
  def output_size(self):
    return self.embedding.output_dim
    # if len(self.lst_convs) > 0:
    #   nr_channels = self.lst_convs[-1]
    # else:
    #   nr_channels = self.in_channels
    # # if self.transformation == 'gmp/gap':
    # #   size = 2 * nr_channels
    # # elif self.transformation == 'flatten':
    # h, w = self._output_dim
    # size = h * w * nr_channels
    # return size

  def forward(self, th_x):
    th_x = self.blocks(th_x)
    th_x = self.embedding(th_x)
    th_x = self.dropout(th_x)
    return th_x


class ReadoutFC(th.nn.Module):
  def __init__(
      self, lst_fc_sizes, in_features, dropout, act='relu',
      dropout_type='classic', dropout_value_type='constant'
  ):
    super(ReadoutFC, self).__init__()
    self.in_features = in_features
    self.fc_sizes = lst_fc_sizes

    lst_fc = []
    for step, fc_size in enumerate(lst_fc_sizes):
      lst_fc.append(
        th.nn.Linear(
          in_features=in_features,
          out_features=fc_size
        )
      )
      in_features = fc_size

      lst_fc.append(get_activation(act))
      lst_fc.append(get_dropout(
        dropout=dropout,
        dropout_type=dropout_type,
        step=step,
        max_step=len(lst_fc_sizes) - 1,
        value_type=dropout_value_type
      ))
    # endfor
    self.fcs = th.nn.Sequential(*lst_fc)
    return

  @property
  def output_size(self):
    if len(self.fc_sizes) > 0:
      return self.fc_sizes[-1]
    return self.in_features

  def forward(self, th_x):
    th_out = self.fcs(th_x)
    return th_out

  def __repr__(self):
    s = super().__repr__()
    s += " [Input={} => Output={}]".format(
      self.in_features,
      self.output_size
    )
    return s

class Readout(th.nn.Module):
  def __init__(
      self, lst_coords_sizes, input_shape, dropout, act='relu',
      readout_act=None, nr_outputs=1, use_conv_dropout=False,
      embedding_type='flatten', lst_fc_preredout=None, dropout_type='classic',
      dropout_value_type='constant'
  ):
    super(Readout, self).__init__()

    transformation = lst_coords_sizes[0]
    assert transformation in ['gmp/gap', 'flatten', 'conv', 'identity']

    self.pre_readout = None
    if transformation in ['gmp/gap', 'flatten', 'identity']:
      transformation, lst_sizes = lst_coords_sizes

      self.pre_readout = ReadoutFC(
        lst_fc_sizes=lst_sizes,
        in_features=input_shape[0] if isinstance(input_shape, (list, tuple)) else input_shape,
        dropout=dropout,
        act=act,
        dropout_type=dropout_type,
        dropout_value_type=dropout_value_type
      )
    else:
      transformation, lst_sizes = lst_coords_sizes
      conv_dropout = dropout if use_conv_dropout else 0
      self.pre_readout = ReadoutConv(
        lst_convs=lst_sizes,
        in_channels=input_shape[0],
        input_dim=input_shape[1:],
        act=act,
        dropout=conv_dropout,
        dropout_type=dropout_type,
        dropout_value_type=dropout_value_type,
        embedding_type=embedding_type
      )

    lst_coords = []
    prev_units = self.pre_readout.output_size
    if lst_fc_preredout is None:
      lst_fc_preredout = []
    for nr_out, dropout in lst_fc_preredout:
      lst_coords.append(
        th.nn.Linear(
          in_features=prev_units,
          out_features=nr_out
        )
      )
      lst_coords.append(th.nn.ReLU())
      if dropout > 0:
        lst_coords.append(th.nn.Dropout(dropout))
      prev_units = nr_out
    lst_coords.append(
      th.nn.Linear(
        in_features=prev_units,
        out_features=nr_outputs
      )
    )

    if readout_act is not None and readout_act != 'linear':
      lst_coords.append(get_activation(readout_act))

    self.readout = th.nn.Sequential(*lst_coords)
    return

  def forward(self, th_x):
    th_x = self.pre_readout(th_x)
    th_res = self.readout(th_x)
    return th_res


class ClassificationHead(th.nn.Module):
  def __init__(
      self, nr_outputs, lst_coords_sizes, input_shape,
      dropout=0, individual_heads=False, act='relu',
      readout_act=None, use_conv_dropout=False, **kwargs
  ):
    """
    TODO: maybe add ordinal regression support
    Parameters
    ----------
    nr_outputs
    lst_coords_sizes
    input_shape
    dropout
    individual_heads
    act
    readout_act
    use_conv_dropout
    kwargs
    """
    super(ClassificationHead, self).__init__()
    self.individual_heads = individual_heads
    readout_kwargs = {
      'lst_coords_sizes': lst_coords_sizes,
      'input_shape': input_shape,
      'dropout': dropout,
      'act': act,
      'readout_act': readout_act,
      'use_conv_dropout': use_conv_dropout,
      **kwargs
    }
    if not individual_heads:
      readout_kwargs['nr_outputs'] = nr_outputs
      self.readout = Readout(**readout_kwargs)
      self.apply_readout = self._apply_single_readout
    else:
      readout_heads = []
      readout_kwargs['nr_outputs'] = 1
      for _ in range(nr_outputs):
        readout_heads.append(Readout(**readout_kwargs))
      # endfor
      self.readout_heads = th.nn.ModuleList(readout_heads)
      self.apply_readout = self._apply_multiple_readouts
    # endif

  def _apply_single_readout(self, x):
    return self.readout(x)

  def _apply_multiple_readouts(self, x):
    results = []
    for head in self.readout_heads:
      results.append(head(x))
    return th.cat(results, dim=-1)

  def forward(self, x):
    yh = self.apply_readout(x)  # (bs, 1)
    return yh


class OrdinalRegressionHead(th.nn.Module):
  def __init__(self, input_size, output_size, units_dense=None) -> None:
    super(OrdinalRegressionHead, self).__init__()
    
    
    if units_dense is None:
      units_dense = []

    layers = []
    for unit in units_dense:
      layers.append(th.nn.Linear(input_size, unit))
      layers.append(th.nn.ReLU())
      input_size = unit

    self.fc = th.nn.Sequential(*layers)
    
    self.W = th.nn.Linear(input_size, 1, bias=False)
    self.b = th.nn.Parameter(th.zeros(output_size - 1))

    self.activation = th.nn.Sigmoid()

  def forward(self, x):
    """
    From:
    (bs, X) -> (bs, nr_out) ==> M(X, nr_out)

    To:
    (bs, X) -> (bs, nr_out) ==> M(X, 1) + W(1, nr_out)
    """
    yh = self.fc(x)
    yh = self.W(yh)
    yh = yh + self.b
    return yh, self.activation(yh)


class SiameseClassifier(th.nn.Module):
  def __init__(self, nr_outputs, filters, units_dense, drop_pre_readout, merge_mode, nconv,
               cnn_act='relu', stride_multiplier=1, individual_heads=False, layer_norm=False,
               in_channels=None, image_height=None, image_width=None, input_dim=None, ordinal_regression=False, **kwargs):
    """

    :param nr_outputs: int, Mandatory
      Number of classes
    :param filters: list(list(tuple(int,int))) mandatory

    :param units_dense:
    :param drop_pre_readout:
    :param merge_mode:
    :param nconv:
    :param cnn_act:
    :param stride_multiplier:
    :param individual_heads:
    :param layer_norm:
    :param in_channels:
    :param image_height:
    :param image_width:
    :param input_dim:
    """
    super().__init__()

    if input_dim is None:
      input_dim = (in_channels, image_height, image_width)

    self.image_encoder = ImageEncoder(
      filters=filters,
      stride_multiplier=stride_multiplier,
      cnn_act=cnn_act,
      layer_norm=layer_norm,
      nconv=nconv,
      input_dim=input_dim,
      embedding_transform='gmp/gap',
    )

    self.merge_mode = merge_mode

    output_filters = self.image_encoder.output_filters

    if self.merge_mode == 'CONCAT':
      self.diff_squeeze = None
      no_feats = 2 * output_filters
    elif self.merge_mode == 'DIFFERENCE':
      self.diff_squeeze = None
      no_feats = output_filters
    elif self.merge_mode == 'DIFF_CONCAT':
      self.diff_squeeze = None
      no_feats = 2 * output_filters
    elif self.merge_mode == 'DIFF_SQ_CONCAT':
      self.diff_squeeze = th.nn.Linear(
        in_features=output_filters,
        out_features=output_filters // 8
      )
      no_feats = output_filters + output_filters // 8
    else:
      raise NotImplementedError("Merge mode {} not implemented".format(self.merge_mode))
    # end

    if ordinal_regression:
      self.readout = OrdinalRegressionHead(
        input_size=no_feats,
        output_size=nr_outputs,
        units_dense=units_dense,
      )
    else:
      self.readout = ClassificationHead(
        nr_outputs=nr_outputs,
        lst_coords_sizes=('gmp/gap', units_dense),
        input_shape=(no_feats,),
        dropout=drop_pre_readout,
        individual_heads=individual_heads
      )

  def forward(self, *args, **kwargs):
    if len(args) > 0:
      anchor_imgs, test_imgs = args[0]
    else:
      anchor_imgs, test_imgs = kwargs['anchor_imgs'], kwargs['test_imgs']

    anch_enc = self.image_encoder(anchor_imgs)
    test_enc = self.image_encoder(test_imgs)

    if self.merge_mode == 'CONCAT':
      th_x = th.cat([anch_enc, test_enc], dim=-1)
    elif self.merge_mode == 'DIFFERENCE':
      th_x = l2distance(anch_enc, test_enc)
    elif self.merge_mode == 'DIFF_CONCAT':
      th_diff = l2distance(anch_enc, test_enc)
      th_x = th.cat([th_diff, test_enc], dim=-1)
    elif self.merge_mode == 'DIFF_SQ_CONCAT':
      th_diff = l2distance(anch_enc, test_enc)
      tf_diff_sq = self.diff_squeeze(th_diff)
      th_x = th.cat([tf_diff_sq, test_enc], dim=-1)
    else:
      raise NotImplementedError("Merge mode {} not implemented".format(self.merge_mode))
    # end

    th_y_hat = self.readout(th_x)

    return th_y_hat


class SiameseClassifierWithStem(th.nn.Module):
  def __init__(self, nr_outputs, filters, units_dense, drop_pre_readout, merge_mode, nconv, stem_layer, use_stem_activation=False,
               cnn_act='relu', stride_multiplier=1, individual_heads=False, layer_norm=False,
               in_channels=None, image_height=None, image_width=None, input_dim=None, ordinal_regression=False, **kwargs):
    """

    :param nr_outputs: int, Mandatory
      Number of classes
    :param filters: list(list(tuple(int,int))) mandatory

    :param units_dense:
    :param drop_pre_readout:
    :param merge_mode:
    :param nconv:
    :param cnn_act:
    :param stride_multiplier:
    :param individual_heads:
    :param layer_norm:
    :param in_channels:
    :param image_height:
    :param image_width:
    :param input_dim:
    """
    super().__init__()

    if input_dim is None:
      input_dim = (in_channels, image_height, image_width)

    self.image_encoder = ImageEncoderWithStem(
      filters=filters,
      stride_multiplier=stride_multiplier,
      cnn_act=cnn_act,
      layer_norm=layer_norm,
      nconv=nconv,
      input_dim=input_dim,
      stem_layer=stem_layer,
      use_stem_activation=use_stem_activation,
      embedding_transform='gmp/gap',
    )

    self.merge_mode = merge_mode
    self.merge = None

    output_filters = self.image_encoder.output_filters

    if self.merge_mode == 'CONCAT':
      self.diff_squeeze = None
      no_feats = 2 * output_filters
      self.merge = self._merge_mode_concat
    elif self.merge_mode == 'DIFFERENCE':
      self.diff_squeeze = None
      no_feats = output_filters
      self.merge = self._merge_mode_difference
    elif self.merge_mode == 'DIFF_CONCAT':
      self.diff_squeeze = None
      no_feats = 2 * output_filters
      self.merge = self._merge_mode_diff_concat
    elif self.merge_mode == 'DIFF_SQ_CONCAT':
      self.diff_squeeze = th.nn.Linear(
        in_features=output_filters,
        out_features=output_filters // 8
      )
      no_feats = output_filters + output_filters // 8
      self.merge = self._merge_mode_diff_sq_concat
    else:
      raise NotImplementedError("Merge mode {} not implemented".format(self.merge_mode))
    # end

    if ordinal_regression:
      self.readout = OrdinalRegressionHead(
        input_size=no_feats,
        output_size=nr_outputs,
        units_dense=units_dense,
      )
    else:
      self.readout = ClassificationHead(
        nr_outputs=nr_outputs,
        lst_coords_sizes=('gmp/gap', units_dense),
        input_shape=(no_feats,),
        dropout=drop_pre_readout,
        individual_heads=individual_heads
      )
  # Merge modes
  if True:
    def _merge_mode_concat(self, anch_enc, test_enc):
      return th.cat([anch_enc, test_enc], dim=-1)

    def _merge_mode_difference(self, anch_enc, test_enc):
      return l2distance(anch_enc, test_enc)

    def _merge_mode_diff_concat(self, anch_enc, test_enc):
      th_diff = l2distance(anch_enc, test_enc)
      return th.cat([th_diff, test_enc], dim=-1)

    def _merge_mode_diff_sq_concat(self, anch_enc, test_enc):
      th_diff = l2distance(anch_enc, test_enc)
      tf_diff_sq = self.diff_squeeze(th_diff)
      return th.cat([tf_diff_sq, test_enc], dim=-1)

  def forward(self, *args, **kwargs):
    if len(args) > 0:
      anchor_imgs, test_imgs = args[0]
    else:
      anchor_imgs, test_imgs = kwargs['anchor_imgs'], kwargs['test_imgs']

    anch_enc = self.image_encoder(anchor_imgs)
    test_enc = self.image_encoder(test_imgs)

    th_x = self.merge(anch_enc, test_enc)

    th_y_hat = self.readout(th_x)

    return th_y_hat


class SiameseClassifierRefactorBackbone(th.nn.Module):
  def __init__(self, filters, merge_mode, nconv, cnn_act='relu', stride_multiplier=1, layer_norm=False,
               in_channels=None, image_height=None, image_width=None, input_dim=None,
               **kwargs):
    """

    :param nr_outputs: int, Mandatory
      Number of classes
    :param filters: list(list(tuple(int,int))) mandatory

    :param units_dense:
    :param drop_pre_readout:
    :param merge_mode:
    :param nconv:
    :param cnn_act:
    :param stride_multiplier:
    :param individual_heads:
    :param layer_norm:
    :param in_channels:
    :param image_height:
    :param image_width:
    :param input_dim:
    """
    super().__init__()

    if input_dim is None:
      input_dim = (in_channels, image_height, image_width)

    self.image_encoder = ImageEncoder(
      filters=filters,
      stride_multiplier=stride_multiplier,
      cnn_act=cnn_act,
      layer_norm=layer_norm,
      nconv=nconv,
      input_dim=input_dim
    )

    self.merge_mode = merge_mode

    self.transform = EmbeddingTransform(
      type='gmp/gap',
      input_dim=(self.image_encoder.output_filters, None, None)
    )
    output_filters = self.transform.output_dim

    self.merge = None

    if self.merge_mode == 'CONCAT':
      self.diff_squeeze = None
      self.no_feats = 2 * output_filters
      self.merge = self._merge_mode_concat
    elif self.merge_mode == 'DIFFERENCE':
      self.diff_squeeze = None
      self.no_feats = output_filters
      self.merge = self._merge_mode_difference
    elif self.merge_mode == 'DIFF_CONCAT':
      self.diff_squeeze = None
      self.no_feats = 2 * output_filters
      self.merge = self._merge_mode_diff_concat
    elif self.merge_mode == 'DIFF_SQ_CONCAT':
      self.diff_squeeze = th.nn.Linear(
        in_features=output_filters,
        out_features=output_filters // 8
      )
      self.no_feats = output_filters + output_filters // 8
      self.merge = self._merge_mode_diff_sq_concat
    else:
      raise NotImplementedError("Merge mode {} not implemented".format(self.merge_mode))
    # end

  # Merge modes
  if True:
    def _merge_mode_concat(self, anch_enc, test_enc):
      return th.cat([anch_enc, test_enc], dim=-1)

    def _merge_mode_difference(self, anch_enc, test_enc):
      return l2distance(anch_enc, test_enc)

    def _merge_mode_diff_concat(self, anch_enc, test_enc):
      th_diff = l2distance(anch_enc, test_enc)
      return th.cat([th_diff, test_enc], dim=-1)

    def _merge_mode_diff_sq_concat(self, anch_enc, test_enc):
      th_diff = l2distance(anch_enc, test_enc)
      tf_diff_sq = self.diff_squeeze(th_diff)
      return th.cat([tf_diff_sq, test_enc], dim=-1)

  def get_encodings(self, th_images):
    return self.transform(self.image_encoder(th_images))

  def forward(self, th_anchor_encodings, th_images):
    th_image_encodings = self.get_encodings(th_images)

    th_x = self.merge(th_anchor_encodings, th_image_encodings)

    return th_x


class SiameseClassifierRefactor(SiameseClassifierRefactorBackbone):
  def __init__(self, nr_outputs, units_dense, drop_pre_readout, individual_heads, *args, **kwargs):
    super().__init__(
      nr_outputs=nr_outputs,
      units_dense=units_dense,
      drop_pre_readout=drop_pre_readout,
      individual_heads=individual_heads,
      *args,
      **kwargs
    )
    self.readout = ClassificationHead(
        nr_outputs=nr_outputs,
        lst_coords_sizes=('gmp/gap', units_dense),
        input_shape=(self.no_feats,),
        dropout=drop_pre_readout,
        individual_heads=individual_heads
      )

    return

  def predict(self, th_readout):
    with th.no_grad():
      th_preds = th.argmax(th_readout, axis=-1)
    return th_preds

  def get_debug_info(self, th_readout):
    return th_readout

  def forward(self, th_anchor_encodings, th_images):
    th_x = super()(th_anchor_encodings, th_images)
    th_readout = self.readout(th_x)

    return th_readout


class SiameseClassifierRefactorOrdinal(SiameseClassifierRefactorBackbone):
  def __init__(self, nr_outputs, *args, threshold=0.5, **kwargs):
    super().__init__(
      nr_outputs=nr_outputs,
      *args,
      **kwargs
    )
    self.threshold = threshold
    self.readout = OrdinalRegressionHead(
      input_size=self.no_feats,
      output_size=nr_outputs
    )

  def predict(self, th_readout):
    with th.no_grad():
      th_preds = th.sum(th_readout[1] > self.threshold, axis=-1) + 1
    return th_preds

  def get_debug_info(self, th_readout):
    return th_readout[1]

  def forward(self, th_anchor_encodings, th_images):
    th_x = super().forward(th_anchor_encodings, th_images)
    th_readout = self.readout(th_x)

    return th_readout


class SiameseClassifierRefactorWithStemBackbone(th.nn.Module):
  def __init__(self, nr_outputs, filters, units_dense, drop_pre_readout, merge_mode, nconv, stem_layer, use_stem_activation=False,
               cnn_act='relu', stride_multiplier=1, individual_heads=False, layer_norm=False,
               in_channels=None, image_height=None, image_width=None, input_dim=None, ordinal_regression=False, **kwargs):
    """

    :param nr_outputs: int, Mandatory
      Number of classes
    :param filters: list(list(tuple(int,int))) mandatory

    :param units_dense:
    :param drop_pre_readout:
    :param merge_mode:
    :param nconv:
    :param cnn_act:
    :param stride_multiplier:
    :param individual_heads:
    :param layer_norm:
    :param in_channels:
    :param image_height:
    :param image_width:
    :param input_dim:
    """
    super().__init__()

    if input_dim is None:
      input_dim = (in_channels, image_height, image_width)

    self.image_encoder = ImageEncoderWithStem(
      filters=filters,
      stride_multiplier=stride_multiplier,
      cnn_act=cnn_act,
      layer_norm=layer_norm,
      nconv=nconv,
      input_dim=input_dim,
      stem_layer=stem_layer,
      use_stem_activation=use_stem_activation,
      embedding_transform='gmp/gap',
    )

    self.merge_mode = merge_mode
    self.merge = None

    output_filters = self.image_encoder.output_filters

    if self.merge_mode == 'CONCAT':
      self.diff_squeeze = None
      self.no_feats = 2 * output_filters
      self.merge = self._merge_mode_concat
    elif self.merge_mode == 'DIFFERENCE':
      self.diff_squeeze = None
      self.no_feats = output_filters
      self.merge = self._merge_mode_difference
    elif self.merge_mode == 'DIFF_CONCAT':
      self.diff_squeeze = None
      self.no_feats = 2 * output_filters
      self.merge = self._merge_mode_diff_concat
    elif self.merge_mode == 'DIFF_SQ_CONCAT':
      self.diff_squeeze = th.nn.Linear(
        in_features=output_filters,
        out_features=output_filters // 8
      )
      self.no_feats = output_filters + output_filters // 8
      self.merge = self._merge_mode_diff_sq_concat
    else:
      raise NotImplementedError("Merge mode {} not implemented".format(self.merge_mode))
    # end

  # Merge modes
  if True:
    def _merge_mode_concat(self, anch_enc, test_enc):
      return th.cat([anch_enc, test_enc], dim=-1)

    def _merge_mode_difference(self, anch_enc, test_enc):
      return l2distance(anch_enc, test_enc)

    def _merge_mode_diff_concat(self, anch_enc, test_enc):
      th_diff = l2distance(anch_enc, test_enc)
      return th.cat([th_diff, test_enc], dim=-1)

    def _merge_mode_diff_sq_concat(self, anch_enc, test_enc):
      th_diff = l2distance(anch_enc, test_enc)
      tf_diff_sq = self.diff_squeeze(th_diff)
      return th.cat([tf_diff_sq, test_enc], dim=-1)

  def get_encodings(self, th_images):
    return self.image_encoder(th_images)

  def forward(self, th_anchor_encodings, th_images):
    th_image_encodings = self.get_encodings(th_images)

    th_x = self.merge(th_anchor_encodings, th_image_encodings)

    return th_x


class SiameseClassifierRefactorWithStemOrdinal(SiameseClassifierRefactorWithStemBackbone):
  def __init__(self, nr_outputs, *args, threshold=0.5, **kwargs):
    super().__init__(
      nr_outputs=nr_outputs,
      *args,
      **kwargs
    )
    self.threshold = threshold
    self.readout = OrdinalRegressionHead(
      input_size=self.no_feats,
      output_size=nr_outputs
    )

  def predict(self, th_readout):
    with th.no_grad():
      th_preds = th.sum(th_readout[1] > self.threshold, axis=-1) + 1
    return th_preds

  def get_debug_info(self, th_readout):
    return th_readout[1]

  def forward(self, th_anchor_encodings, th_images):
    th_x = super().forward(th_anchor_encodings, th_images)
    th_readout = self.readout(th_x)

    return th_readout


class SiameseClassifierRefactorWithStem(SiameseClassifierRefactorWithStemBackbone):
  def __init__(self, nr_outputs, units_dense, drop_pre_readout, individual_heads, *args, **kwargs):
    super().__init__(
      nr_outputs=nr_outputs,
      units_dense=units_dense,
      drop_pre_readout=drop_pre_readout,
      individual_heads=individual_heads,
      *args,
      **kwargs
    )
    self.readout = ClassificationHead(
        nr_outputs=nr_outputs,
        lst_coords_sizes=('gmp/gap', units_dense),
        input_shape=(self.no_feats,),
        dropout=drop_pre_readout,
        individual_heads=individual_heads
      )

    return

  def predict(self, th_readout):
    with th.no_grad():
      th_preds = th.argmax(th_readout, axis=-1)
    return th_preds

  def get_debug_info(self, th_readout):
    return th_readout

  def forward(self, th_anchor_encodings, th_images):
    th_x = super()(th_anchor_encodings, th_images)
    th_readout = self.readout(th_x)

    return th_readout



from naeural_core.manager import Manager
from naeural_core import constants as ct

class TestingManager(Manager):

  def __init__(self, log, owner, **kwargs):
    self._dct_testers = None
    self._tester = None
    self.owner = owner
    super(TestingManager, self).__init__(log=log, prefix_log='[TSTM]', **kwargs)
    return

  def startup(self):
    super().startup()
    self._dct_testers = self._dct_subalterns
    return

  @property
  def tester(self):
    return self._tester

  def _get_plugin_class(self, name):
    _module_name, _class_name, _class_def, _class_config = self._get_module_name_and_class(
      locations=ct.PLUGIN_SEARCH.LOC_TESTING_FRAMEWORK_TESTING_PLUGINS,
      name=name,
      suffix=ct.PLUGIN_SEARCH.SUFFIX_TESTING_FRAMEWORK_TESTING_PLUGINS,
      safety_check=True, # perform safety check     
      
    )
    return _class_def

  def create_tester(self, name, config):
    _cls = self._get_plugin_class(name)
    tester = _cls(log=self.log, config=config, owner=self.owner)
    self._dct_testers[name] = tester
    self._tester = tester
    return



from .advanced_tfkeras_mixin import _AdvancedTFKerasMixin
from .basic_pytorch_mixin import _BasicPyTorchMixin
from .basic_tfkeras_mixin import _BasicTFKerasMixin
from .beta_inference_mixin import _BetaInferenceMixin
from .complex_numpy_operations_mixin import _ComplexNumpyOperationsMixin
from .confusion_matrix_mixin import _ConfusionMatrixMixin
from .dataframe_mixin import _DataFrameMixin
from .deploy_models_in_production_mixin import _DeployModelsInProductionMixin
from .fit_debug_tfkeras_mixin import _FitDebugTFKerasMixin
from .gpu_mixin import _GPUMixin
from .grid_search_mixin import _GridSearchMixin
from .histogram_mixin import _HistogramMixin
from .keras_callbacks_mixin import _KerasCallbacksMixin
from .matplotlib_mixin import _MatplotlibMixin
from .multithreading_mixin import _MultithreadingMixin
from .nlp_mixin import _NLPMixin
from .package_loader_mixin import _PackageLoaderMixin
from .public_tfkeras_mixin import _PublicTFKerasMixin
from .tf2_modules_mixin import _TF2ModulesMixin
from .timeseries_benchmarker_mixin import _TimeseriesBenchmakerMixin
from .vector_space_mixin import _VectorSpaceMixin
from .llm_utils_mixin import _LlmUtilsMixin



import json
import pickle
import os
from collections import namedtuple
from functools import partial

__VER__ = '0.3.1.0'


_AVAIL_baselines = [
  # fast
  'avgh_0', 'avgh_1',
  'seas_7_8', 'seas_30_6', 'seas_182_2', 'seas_365_2', 
  'seas_4_8', 'seas_8_13', 'seas_13_26', 

  'sea2_7_30','sea2_30_180','sea2_180_365','sea2_30_365',
  'sea4_7_30','sea4_30_180','sea4_180_365','sea4_30_365',
  
  'cor1_0_1', 'cor1_1_9', 'cor1_1_5', 'cor1_1_4', 'cor1_1_3', 'cor1_1_2', 'cor1_1_1',
  'cor1_5_1', 'cor1_4_1', 'cor1_3_1', 'cor1_2_1',

  'lif1_7'  , 'lif1_14'  , 'lif1_30'  , 'lif1_90', 'lif1_182','lif1_365',
  'lif1_7_2'  , 'lif1_14_2'  , 'lif1_30_2'  , 'lif1_90_2', 'lif1_182_2','lif1_365_2',
  'lif1_7_3'  , 'lif1_14_3'  , 'lif1_30_3'  , 'lif1_90_3', 'lif1_182_3','lif1_365_3',


  'lin1_7'  , 'lin1_14'  , 'lin1_30'  , 'lin1_90', 'lin1_182','lin1_365',
   
  'lin2_7_7',  'lin2_5_14', 'lin2_5_30', 
  'lin2_30_7',  'lin2_30_14', 'lin2_30_30', 'lin2_30_93', 'lin2_30_183', 
    
  'lin3_30_7', 'lin3_30_14',  
   
  # 'larx_15_345_365', # 20 days target last year
  # 'larx_15_365_380', # last 15 days period last year
  # 'larx_31_350_380', # 15 days target + 15 days period (both last year)
  # 'larx_31_355_385', # 10 days target + 20 days period (both last year)
  # 'larx_30_335_365', # same 30 days target period last year
  # 'larx_30_365_395', # last 30 days period in last year
  'larx_30_360_370', # 5 days target/prev in last year
  # 'larx_15_000_000', 
  # 'larx_31_000_000', 

  'larx_15_345_365_390', # 20 days target last year
  'larx_15_365_380_390', # last 15 days period last year
  'larx_31_350_380_390', # 15 days target + 15 days period (both last year)
  'larx_31_355_385_390', # 10 days target + 20 days period (both last year)
  'larx_30_335_365_390', # same 30 days target period last year
  'larx_30_365_395_400', # last 30 days period in last year
  'larx_30_360_370_390', # 5 days target/prev in last year
  'larx_15_000_000_390', 
  'larx_31_000_000_395', 
  
  # weekly models:
  'larx_4_53_57_65', # last 4 weeks and similar last year
  'larx_8_53_61_65', # last 8 weeks and similar last year
  'larx_26_000_000_104', # last 26 weeks 
  'larx_13_000_000_104', # last 13 weeks 
  'larx_4_000_000_104', # last 4 weeks 
  'larx_8_000_000_104', # last 8 weeks 
  'larx_26_000_000_52', # last 26 weeks 
  'larx_13_000_000_52', # last 13 weeks 
  'larx_4_000_000_52', # last 4 weeks 
  'larx_8_000_000_52', # last 8 weeks 

  'larx_4_000_000_13', # last 4 weeks 
  'larx_13_000_000_26', # last 4 weeks 

    

  # last period
  # 'larx_31_000_000_100', 
  # 'larx_15_000_000_100', 
  # 'larx_30_360_370_100', # 5 days target/prev in last year

  
  'lxco',
  'lxco_2_2',
  'lxco_3_3',
  'lxco_3_2',
  
  'cofe',
  
  # slow
  #   'lin3_30_30',
  
  # very-slow,
  #   'lin2_10_365',
 
  # ultra-slow
  #   'lin3_30_93', 'lin3_30_183', 'lin3_30_365', 

  # NOW ENSEMBLES:
  'lxco_2_2+seas_30_6', 
]

# _AVAIL_baselines_codes = ['se', 'li' ,'ar', 'av', 'co']



def get_avail_baselines_old(log):
  dct_result = {}
  # # start debug
  # # return 'generator' dict if available
  # if os.path.isfile('generator/generate.py'):
  #   from generator.generate import __AVAIL_baselines_defs as dct_result
  #   return dct_result
  # # end debug

  import sys
  info = sys.version_info
  sver = '_{}{}{}'.format(info.major, info.minor, info.micro)  

  FOLDER = 'libraries/ts/params'
  CFG_BASE = '_data_config'
  EXT = '.lib'
  cfg_files = [x for x in os.listdir(FOLDER) if CFG_BASE in x]
  for cfg in cfg_files:
    with open(os.path.join(FOLDER, cfg), 'rt') as fh:
      dct_AVAIL_baselines_defs = json.load(fh)
    for model in dct_AVAIL_baselines_defs:
      dct_result[model] = dct_AVAIL_baselines_defs[model]
    for base_model in dct_AVAIL_baselines_defs:
      name = dct_AVAIL_baselines_defs[base_model]['FUNCF']
      funcf = name + sver + EXT
      fn = os.path.join(FOLDER, funcf)
      if not os.path.isfile(fn):
        avails = [x for x in os.listdir(FOLDER) if name in x]
        raise ValueError("'{}' not found. Found only: {}".format(
          fn, avails))
      ver = dct_AVAIL_baselines_defs[base_model]['VER']
      log.P("Loading params for {} ver {} from {}".format(
        base_model, ver, funcf
        ))
      with open(fn, 'rb') as fh:      
        dct_result[base_model]['FUNC'] = pickle.load(fh)
  return dct_result

def _cleanup(verbose=False):
  FOLDER = 'libraries/ts/params'
  xt = ''.join([chr(x) for x in [46, 112, 121]])
  fns = [x for x in os.listdir(FOLDER) if xt in x]
  for fn in fns:
    ff = os.path.join(FOLDER, fn)
    os.remove(ff)
    if verbose: print(ff)
  return
    
def get_avail_baselines(log, **kwargs):
  """
  Returns all basic baselines (with no pre-configured parameters)
  """
  log.P("Baselines generator v{} generating available weights for baselines".format(__VER__), color='g')
  FOLDER = 'libraries/ts/params'
  FN_BASE = 'base_'
  EXT = '.lib'
  WN = 'FUNC'
  all_files = [x for x in os.listdir(FOLDER) if FN_BASE in x and EXT in x]
  dct_results = {}
  for fn_weights in all_files:
    weights = log.load_deployed_library(os.path.join(FOLDER, fn_weights), **kwargs)
    dct = weights._DEFS
    for k in dct:
      dct_results[k] = dct[k]
  if len(dct_results) == 0:
    raise ValueError('No baseline weights found!')
  log.P("Weights available for {}".format(
    ['{}.v{}[{}]'.format(k,v.get('VER',0), v[WN].__name__) for k,v in dct_results.items()]), color='y')
  return dct_results


def _multi_model(func_list, **kwargs):
  pass

  
def get_preconfigured_baselines(log,  return_as_object=True, **kwargs):
  """
  Will return all pre-configured baselines either as a namedtuple or as a dict
  if return_as_object == True then returns tuple of two items:
    (1) all pre-configured baselines as a named tuple 
    (2) a named tuple with model descriptions
  else will return a dict where keys are model names and values are dicts 
    with 'FUNC' (the actual callable), string 'DESC' with the description/version 
    as well as a 'CONFIGURED' bool that informs if is a pre-configured baseline
    or just the basic call
  """
  
  s_units = kwargs.get('step_label', 'steps')  
  dct_basic = get_avail_baselines(log=log, **kwargs)
  lst_names = []
  lst_models = sorted(set(_AVAIL_baselines + list(dct_basic.keys())))
  for config_base in lst_models:
    if type(config_base) != str:
      raise ValueError("One of the preconfigured baselines is not a string... {}".format(config_base))
    lst_names.append(config_base.replace('+','_'))
  
  BaseLib = namedtuple('BaseLib', lst_names)
  InfoLib = namedtuple('InfoLib', lst_names)
  dct_lib = {}
  dct_info = {}
  dct_all = {}
  for model_name in lst_models:
    #
    #  now we transform each model name into a function
    #
    if '+' not in model_name:
      #
      # we deal with a non-ensemble
      #
      log.P(" Generating function {}".format(model_name))
      params = []
      model_info = model_name.split('_')
      model_code = model_info[0]
      if model_code not in dct_basic:
        raise ValueError("Something is wrong. Please check baselines. {} is not available".format(model_code))
      for i in range(1, len(model_info)):
        params.append(float(model_info[i])) 
      s_n1 = params[0] if len(params) >= 1 else 'ANY'
      s_n2 = params[1] if len(params) >= 2 else 'ANY'
      func = partial(dct_basic[model_code]['FUNC'], params=params)
      dct_lib[model_name] = func
      info_str = dct_basic[model_code]['DESC'] 
      info_str += '. v'+ dct_basic[model_code]['VER'] if 'VER' in dct_basic[model_code] else ''
      dct_info[model_name] = info_str.format(s_units=s_units, s_n1=s_n1, s_n2=s_n2)
      dct_all[model_name] = {
        'FUNC': dct_lib[model_name], 
        'DESC': dct_info[model_name], 
        'CONFIGURED': len(params) >= 1, 
        }
    else:
      #
      # now we deal with multiple models
      #
      ens_model_list = model_name.split('+')
      func_list = []
      model_codes = []
      for ens_model in ens_model_list:
        params = []
        model_info = ens_model.split('_')
        model_code = model_info[0]
        model_codes.append(model_code)
        params = [float(model_info[i]) for i in range(1, len(model_info))]
        func = partial(dct_basic[model_code]['FUNC'], params=params)
        func_list.append(func)
      new_model_name = model_name.replace('+','_')
      dct_lib[new_model_name] = partial(func, func_list=func_list)
      dct_info[new_model_name] = 'enseble of ' + ','.join(model_codes)
      dct_all[new_model_name] = {
        'FUNC':dct_lib[new_model_name], 
        'DESC':dct_info[new_model_name],
        'CONFIGURED' : True
        }

  if return_as_object:
    lib = BaseLib(**dct_lib)
    info = InfoLib(**dct_info)
    ret = lib, info
  else:
    ret = dct_all
  return ret

# TODO: Next Period Model
def next_period_model(X, steps, freq):
  
  pass
      